"""Runway Module definition model."""

from __future__ import annotations

from pathlib import Path
from typing import Annotated, Any

from pydantic import ConfigDict, Field, ValidationInfo, field_validator, model_validator

from .. import utils
from ..base import ConfigProperty
from ._type_defs import (
    RunwayEnvironmentsUnresolvedType,
    RunwayEnvVarsUnresolvedType,
    RunwayModuleTypeTypeDef,
)


class RunwayModuleDefinitionModel(ConfigProperty):
    """Defines a directory containing IaC, parameters, and options for the module type."""

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "description": "Defines a directory containing IaC, "
            "the parameters to pass in during execution, "
            "and any applicable options for the module type.",
        },
        title="Runway Module Definition",
        use_enum_values=True,
        validate_default=True,
        validate_assignment=True,
    )

    class_path: Annotated[
        str | None,
        Field(
            description="Import path to a custom Runway module class. (supports lookups)",
        ),
    ] = None
    """Import path to a custom Runway module class. (supports lookups)"""

    env_vars: Annotated[
        RunwayEnvVarsUnresolvedType,
        Field(
            title="Environment Variables",
            description="Additional variables to add to the environment when "
            "processing the deployment. (supports lookups)",
            examples=[
                "${var env_vars.${env DEPLOY_ENVIRONMENT}}",
                {
                    "EXAMPLE_VARIABLE": "value",
                    "KUBECONFIG": [".kube", "${env DEPLOY_ENVIRONMENT}", "config"],
                },
            ],
        ),
    ] = {}
    """Additional variables to add to the environment when processing the deployment. (supports lookups)"""

    environments: Annotated[
        RunwayEnvironmentsUnresolvedType,
        Field(
            description="Explicitly enable/disable the deployment for a specific "
            "deploy environment, AWS Account ID, and AWS Region combination. "
            "Can also be set as a static boolean value. (supports lookups)",
            examples=[
                "${var envs.${env DEPLOY_ENVIRONMENT}}",
                {"dev": "123456789012", "prod": "us-east-1"},
                {"dev": True, "prod": False},
                {"dev": ["us-east-1"], "prod": ["us-west-2", "ca-central-1"]},
                {
                    "dev": ["123456789012/us-east-1", "123456789012/us-west-2"],
                    "prod": ["234567890123/us-east-1", "234567890123/us-west-2"],
                },
            ],
        ),
    ] = {}
    """Explicitly enable/disable the deployment for a specific deploy environment,
    AWS Account ID, and AWS Region combination.
    Can also be set as a static boolean value. (supports lookups)

    """

    name: Annotated[
        str,
        Field(
            description="The name of the module to be displayed in logs and the "
            "interactive selection menu."
        ),
    ] = "undefined"
    """The name of the module to be displayed in logs and the interactive selection menu."""

    options: Annotated[
        dict[str, Any] | str, Field(description="Module type specific options. (supports lookups)")
    ] = {}
    """Module type specific options. (supports lookups)"""

    parameters: Annotated[
        dict[str, Any] | str,
        Field(
            description="Used to pass variable values to modules in place of an "
            "environment configuration file. (supports lookups)",
            examples=[
                {"namespace": "example-${env DEPLOY_ENVIRONMENT}"},
                "${var sampleapp.parameters.${env DEPLOY_ENVIRONMENT}}",
            ],
        ),
    ] = {}
    """Used to pass variable values to modules in place of an environment configuration file. (supports lookups)"""

    path: Annotated[
        Path | str | None,
        Field(
            description="Directory (relative to the Runway config file) containing IaC. "
            "(supports lookups)",
            examples=["./", "sampleapp-${env DEPLOY_ENVIRONMENT}.cfn", "sampleapp.sls"],
        ),
    ] = None
    """Directory (relative to the Runway config file) containing IaC. (supports lookups)"""

    tags: Annotated[
        list[str],
        Field(
            description="Array of values to categorize the module which can be used "
            "with the CLI to quickly select a group of modules. "
            "This field is only used by the `--tag` CLI option.",
            examples=[["type:network", "app:sampleapp"]],
        ),
    ] = []
    """Array of values to categorize the module which can be used with the CLI to
    quickly select a group of modules.

    This field is only used by the ``--tag`` CLI option.

    """

    type: Annotated[
        RunwayModuleTypeTypeDef | None,
        Field(
            description="Explicitly define the module type. If not provided, this will be inferred."
        ),
    ] = None
    """Explicitly define the module type. If not provided, this will be inferred."""

    # needs to be last
    parallel: Annotated[
        list[RunwayModuleDefinitionModel],
        Field(
            description="Array of module definitions that can be executed asynchronously. "
            "Incompatible with class_path, path, and type.",
            examples=[[{"path": "sampleapp-01.cfn"}, {"path": "sampleapp-02.cfn"}]],
        ),
    ] = []
    """List of module definitions that can be executed asynchronously.
    Incompatible with class_path, path, and type.

    """

    @model_validator(mode="before")
    @classmethod
    def _validate_name(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate module name."""
        if "name" in values:
            return values
        if "parallel" in values:
            values["name"] = "parallel_parent"
            return values
        if "path" in values and values:
            values["name"] = Path(values["path"]).resolve().name
            return values
        return values

    @model_validator(mode="before")
    @classmethod
    def _validate_path(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate path and sets a default value if needed."""
        if not values.get("path") and not values.get("parallel"):
            values["path"] = Path.cwd()
        return values

    @field_validator("parallel", mode="before")
    @classmethod
    def _validate_parallel(
        cls, v: list[dict[str, Any] | str], info: ValidationInfo
    ) -> list[dict[str, Any]]:
        """Validate parallel."""
        if v and info.data.get("path"):
            raise ValueError("only one of parallel or path can be defined")
        result: list[dict[str, Any]] = []
        for mod in v:
            if isinstance(mod, str):
                result.append({"path": mod})
            else:
                result.append(mod)
        return result

    # TODO(kyle): add regex to schema
    _validate_string_is_lookup = field_validator(
        "env_vars", "environments", "options", "parameters", mode="before"
    )(utils.validate_string_is_lookup)



"""Utility functions."""

from __future__ import annotations

import hashlib
import importlib
import json
import logging
import os
import platform
import re
import stat
import sys
from collections.abc import Iterable, Iterator, MutableMapping
from contextlib import AbstractContextManager, contextmanager
from functools import cached_property  # noqa: F401  # TODO (kyle): remove in next major release
from pathlib import Path
from subprocess import check_call
from typing import TYPE_CHECKING, Any, Callable, cast, overload

import yaml
from pydantic import BaseModel as _BaseModel

# make this importable without defining __all__ yet.
# more things need to be moved of this file before starting an explicit __all__.
from . import pydantic_validators  # noqa: F401
from ._file_hash import FileHash  # noqa: F401
from ._json_encoder import JsonEncoder  # noqa: F401
from ._version import Version  # noqa: F401

if TYPE_CHECKING:
    from types import TracebackType

    from mypy_boto3_cloudformation.type_defs import OutputTypeDef
    from typing_extensions import Literal

    from ..compat import Self

AWS_ENV_VARS = ("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN")
DOC_SITE = "https://runway.readthedocs.io"
EMBEDDED_LIB_PATH = str(Path(__file__).resolve().parent / "embedded")

LOGGER = logging.getLogger(__name__)


class BaseModel(_BaseModel):
    """Base class for Runway models."""

    def get(self, name: str, default: Any = None) -> Any:
        """Safely get the value of an attribute.

        Args:
            name: Attribute name to return the value for.
            default: Value to return if attribute is not found.

        """
        return getattr(self, name, default)

    def __contains__(self, name: object) -> bool:
        """Implement evaluation of 'in' conditional.

        Args:
            name: The name to check for existence in the model.

        """
        if name in self.__dict__:
            return True
        # extra files are no longer added to __dict__, they are placed in `model_extra`
        return bool(self.model_extra and name in self.model_extra)

    def __getitem__(self, name: str) -> Any:
        """Implement evaluation of self[name].

        Args:
            name: Attribute name to return the value for.

        Returns:
            The value associated with the provided name/attribute name.

        Raises:
            AttributeError: If attribute does not exist on this object.

        """
        return getattr(self, name)

    def __setitem__(self, name: str, value: Any) -> None:
        """Implement item assignment (e.g. ``self[name] = value``).

        Args:
            name: Attribute name to set.
            value: Value to assign to the attribute.

        """
        super().__setattr__(name, value)


class MutableMap(MutableMapping[str, Any]):
    """Base class for mutable map objects."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialize class.

        Provided ``kwargs`` are added to the object as attributes.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                print(obj.__dict__)
                # {'key': 'value'}

        """
        for key, value in kwargs.items():
            if isinstance(value, dict):
                setattr(self, key, MutableMap(**value))
            else:
                setattr(self, key, value)
        if kwargs:
            self._found_queries = MutableMap()

    @property
    def data(self) -> dict[str, Any]:
        """Sanitized output of __dict__.

        Removes anything that starts with ``_``.

        """
        result: dict[str, Any] = {}
        for key, val in self.__dict__.items():
            if key.startswith("_"):
                continue
            result[key] = val.data if isinstance(val, MutableMap) else val
        return result

    def clear_found_cache(self) -> None:
        """Clear _found_cache."""
        for val in self.__dict__.values():
            if isinstance(val, MutableMap):
                val.clear_found_cache()
        if hasattr(self, "_found_queries"):
            self._found_queries.clear()

    def find(self, query: str, default: Any = None, ignore_cache: bool = False) -> Any:
        """Find a value in the map.

        Previously found queries are cached to increase search speed. The
        cached value should only be used if values are not expected to change.

        Args:
            query: A period delimited string that is split to search for
                nested values
            default: The value to return if the query was unsuccessful.
            ignore_cache: Ignore cached value.

        """
        if not hasattr(self, "_found_queries"):
            # if not created from kwargs, this attr won't exist yet
            # this is done to prevent endless recursion
            self._found_queries = MutableMap()

        if not ignore_cache:
            cached_result = self._found_queries.get(query, None)
            if cached_result:
                return cached_result

        split_query = query.split(".")

        if len(split_query) == 1:
            result = self.get(split_query[0], default)
            if result != default:
                self._found_queries[split_query[0]] = result
            return result

        nested_value = self.get(split_query[0])

        if not nested_value:
            if self.get(query):
                return self.get(query)
            return default

        nested_value = nested_value.find(split_query[1])

        try:
            nested_value = self[split_query[0]].find(
                query=".".join(split_query[1:]),
                default=default,
                ignore_cache=ignore_cache,
            )
            if nested_value != default:
                self._found_queries[query] = nested_value
            return nested_value
        except (AttributeError, KeyError):
            return default

    def get(self, key: str, default: Any = None) -> Any:
        """Implement evaluation of self.get.

        Args:
            key: Attribute name to return the value for.
            default: Value to return if attribute is not found.

        """
        return getattr(self, key, default)

    def __bool__(self) -> bool:
        """Implement evaluation of instances as a bool."""
        return bool(self.data)

    def __contains__(self, value: Any) -> bool:
        """Implement evaluation of 'in' conditional."""
        return value in self.data

    def __getitem__(self, key: str) -> Any:
        """Implement evaluation of self[key].

        Args:
            key: Attribute name to return the value for.

        Returns:
            The value associated with the provided key/attribute name.

        Raises:
            AttributeError: If attribute does not exist on this object.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                print(obj['key'])
                # value

        """
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any) -> None:
        """Implement assignment to self[key].

        Args:
            key: Attribute name to associate with a value.
            value: Value of a key/attribute.

        Example:
            .. codeblock: python

                obj = MutableMap()
                obj['key'] = 'value'
                print(obj['key'])
                # value

        """
        if isinstance(value, dict):
            setattr(self, key, MutableMap(**value))
        else:
            setattr(self, key, value)

    def __delitem__(self, key: str) -> None:
        """Implement deletion of self[key].

        Args:
            key: Attribute name to remove from the object.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                del obj['key']
                print(obj.__dict__)
                # {}

        """
        delattr(self, key)

    def __len__(self) -> int:
        """Implement the built-in function len().

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                print(len(obj))
                # 1

        """
        return len(self.__dict__)

    def __iter__(self) -> Iterator[Any]:
        """Return iterator object that can iterate over all attributes.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                for k, v in obj.items():
                    print(f'{key}: {value}')
                # key: value

        """
        return iter(self.__dict__)

    def __str__(self) -> str:
        """Return string representation of the object."""
        return json.dumps(self.data, default=json_serial)


class SafeHaven(AbstractContextManager["SafeHaven"]):
    """Context manager that caches and resets important values on exit.

    Caches and resets os.environ, sys.argv, sys.modules, and sys.path.

    """

    def __init__(
        self,
        argv: Iterable[str] | None = None,
        environ: dict[str, str] | None = None,
        sys_modules_exclude: Iterable[str] | None = None,
        sys_path: Iterable[str] | None = None,
    ) -> None:
        """Instantiate class.

        Args:
            argv: Override the value of sys.argv.
            environ: Update os.environ.
            sys_modules_exclude: A list of modules to exclude when reverting
                sys.modules to its previous state.
                These are checked as ``module.startswith(name)``.
            sys_path: Override the value of sys.path.

        """
        self.__os_environ = dict(os.environ)
        self.__sys_argv = list(sys.argv)
        # deepcopy can't pickle sys.modules and dict()/.copy() are not safe
        self.__sys_modules = {}
        for k, v in sys.modules.items():
            self.__sys_modules[k] = v
        self.__sys_path = list(sys.path)
        # more informative origin for log statements
        self.logger = logging.getLogger("runway." + self.__class__.__name__)
        self.sys_modules_exclude: set[str] = (
            set(sys_modules_exclude) if sys_modules_exclude else set()
        )
        self.sys_modules_exclude.add("runway")

        if isinstance(argv, list):
            sys.argv = argv
        if isinstance(environ, dict):
            os.environ.update(environ)
        if isinstance(sys_path, list):
            sys.path = sys_path

    def reset_all(self) -> None:
        """Reset all values cached by this context manager."""
        self.logger.debug("resetting all managed values...")
        self.reset_os_environ()
        self.reset_sys_argv()
        self.reset_sys_modules()
        self.reset_sys_path()

    def reset_os_environ(self) -> None:
        """Reset the value of os.environ."""
        self.logger.debug("resetting os.environ: %s", json.dumps(self.__os_environ))
        os.environ.clear()
        os.environ.update(self.__os_environ)

    def reset_sys_argv(self) -> None:
        """Reset the value of sys.argv."""
        self.logger.debug("resetting sys.argv: %s", json.dumps(self.__sys_argv))
        sys.argv = self.__sys_argv

    def reset_sys_modules(self) -> None:
        """Reset the value of sys.modules."""
        self.logger.debug("resetting sys.modules...")
        # sys.modules can be manipulated to force reloading modules but,
        # replacing it outright does not work as expected
        for module in list(sys.modules.keys()):
            if module not in self.__sys_modules and not any(
                module.startswith(n) for n in self.sys_modules_exclude
            ):
                self.logger.debug(
                    'removed sys.module: {"%s": "%s"}', module, sys.modules.pop(module)
                )

    def reset_sys_path(self) -> None:
        """Reset the value of sys.path."""
        self.logger.debug("resetting sys.path: %s", json.dumps(self.__sys_path))
        sys.path = self.__sys_path

    def __enter__(self) -> Self:
        """Enter the context manager.

        Returns:
            SafeHaven: Instance of the context manager.

        """
        self.logger.debug("entering a safe haven...")
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit the context manager."""
        self.logger.debug("leaving the safe haven...")
        self.reset_all()


# TODO (kyle): remove after https://github.com/yaml/pyyaml/issues/234 is resolved
class YamlDumper(yaml.Dumper):
    """Custom YAML Dumper.

    This Dumper allows for YAML to be output to follow YAML spec 1.2,
    example 2.3 of collections (2.1). This provides an output that is more
    humanreadable and complies with yamllint.

    Example:
        >>> print(yaml.dump({'key': ['val1', 'val2']}, Dumper=YamlDumper))

    Note:
        YAML 1.2 Specification: https://yaml.org/spec/1.2/spec.html
        used for reference.

    """

    def increase_indent(self, flow: bool = False, indentless: bool = False) -> None:  # noqa: ARG002
        """Override parent method."""
        return super().increase_indent(flow, False)


@contextmanager
def argv(*args: str) -> Iterator[None]:
    """Context manager for temporarily changing sys.argv."""
    original_argv = sys.argv.copy()
    try:
        sys.argv = list(args)  # convert tuple to list
        yield
    finally:
        # always restore original value
        sys.argv = original_argv


@contextmanager
def change_dir(newdir: Path | str) -> Iterator[None]:
    """Change directory.

    Adapted from http://stackoverflow.com/a/24176022

    Args:
        newdir: Path to change directory into.

    """
    prevdir = Path.cwd().resolve()
    if isinstance(newdir, str):
        newdir = Path(newdir)
    os.chdir(newdir.resolve())
    try:
        yield
    finally:
        os.chdir(prevdir)


def ensure_file_is_executable(path: str) -> None:
    """Exit if file is not executable.

    Args:
        path: Path to file.

    Raises:
        SystemExit: file is not executable.

    """
    if platform.system() != "Windows" and (not stat.S_IXUSR & Path(path).stat()[stat.ST_MODE]):
        print(f"Error: File {path} is not executable")  # noqa: T201
        sys.exit(1)


def ensure_string(value: Any) -> str:
    """Ensure value is a string."""
    if isinstance(value, str):
        return value
    if isinstance(value, bytes):
        return value.decode()
    raise TypeError(f"Expected str or bytes but received {type(value)}")


@contextmanager
def environ(env: dict[str, str] | None = None, **kwargs: str) -> Iterator[None]:
    """Context manager for temporarily changing os.environ.

    The original value of os.environ is restored upon exit.

    Args:
        env: Dictionary to use when updating os.environ.
        **kwargs: Arbitrary keyword arguments.

    """
    env = env or {}
    env.update(kwargs)

    original_env = dict(os.environ)
    os.environ.update(env)

    try:
        yield
    finally:
        # always restore original values
        os.environ.clear()
        os.environ.update(original_env)


def json_serial(obj: Any) -> Any:
    """JSON serializer for objects not serializable by default json code."""
    if isinstance(obj, MutableMap):
        return obj.data
    raise TypeError(f"Type {type(obj)} not serializable")


def load_object_from_string(fqcn: str, try_reload: bool = False) -> type | Callable[..., Any]:
    """Convert "." delimited strings to a python object.

    Args:
        fqcn: A "." delimited string representing the full path to an
            object (function, class, variable) inside a module
        try_reload: Try to reload the module so any global variables
            set within the file during import are reloaded. This only applies
            to modules that are already imported and are not builtin.

    Returns:
        Object being imported from the provided path.

    Example::

        load_object_from_string("os.path.basename")
        load_object_from_string("logging.Logger")
        load_object_from_string("LocalClassName")

    """
    module_path = "__main__"
    object_name = fqcn
    if "." in object_name:
        module_path, object_name = fqcn.rsplit(".", 1)
        if (
            try_reload
            and sys.modules.get(module_path)
            and module_path.split(".")[0] not in sys.builtin_module_names  # skip builtins
        ):
            importlib.reload(sys.modules[module_path])
        else:
            importlib.import_module(module_path)
    return getattr(sys.modules[module_path], object_name)


@overload
def merge_dicts(
    dict1: dict[Any, Any], dict2: dict[Any, Any], deep_merge: bool = ...
) -> dict[str, Any]: ...


@overload
def merge_dicts(dict1: list[Any], dict2: list[Any], deep_merge: bool = ...) -> list[Any]: ...


def merge_dicts(
    dict1: dict[Any, Any] | list[Any],
    dict2: dict[Any, Any] | list[Any],
    deep_merge: bool = True,
) -> dict[Any, Any] | list[Any]:
    """Merge dict2 into dict1."""
    if deep_merge:
        if isinstance(dict1, list) and isinstance(dict2, list):
            return dict1 + dict2

        if not isinstance(dict1, dict) or not isinstance(dict2, dict):
            return dict2

        for key in dict2:
            dict1[key] = merge_dicts(dict1[key], dict2[key], True) if key in dict1 else dict2[key]
        return dict1
    if isinstance(dict1, dict) and isinstance(dict2, dict):
        dict3 = dict1.copy()
        dict3.update(dict2)
        return dict3
    raise ValueError(f"values of type {type(dict1)} and {type(dict2)} must be type dict")


def snake_case_to_kebab_case(value: str) -> str:
    """Convert snake_case to kebab-case.

    Args:
        value: The string value to convert.

    """
    return value.replace("_", "-")


def extract_boto_args_from_env(env_vars: dict[str, str]) -> dict[str, str]:
    """Return boto3 client args dict with environment creds."""
    return {
        i: env_vars[i.upper()]
        for i in ["aws_access_key_id", "aws_secret_access_key", "aws_session_token"]
        if env_vars.get(i.upper(), "")
    }


def flatten_path_lists(env_dict: dict[str, Any], env_root: str | None = None) -> dict[str, Any]:
    """Join paths in environment dict down to strings."""
    for key, val in env_dict.items():
        # Lists are presumed to be path components and will be turned back
        # to strings
        if isinstance(val, list):
            env_dict[key] = (
                Path(env_root).joinpath(*cast("list[str]", val))
                if (env_root and not Path(*cast("list[str]", val)).is_absolute())
                else Path(*cast("list[str]", val))
            )
    return env_dict


def merge_nested_environment_dicts(
    env_dicts: dict[str, Any],
    env_name: str | None = None,
    env_root: str | None = None,
) -> dict[str, Any]:
    """Return single-level dictionary from dictionary of dictionaries."""
    # If the provided dictionary is just a single "level" (no nested
    # environments), it applies to all environments
    if all(isinstance(val, (str, list)) for (_key, val) in env_dicts.items()):
        return flatten_path_lists(env_dicts, env_root)

    if env_name is None:
        if env_dicts.get("*"):
            return flatten_path_lists(env_dicts.get("*", {}), env_root)
        return {}

    if not env_dicts.get("*") and not env_dicts.get(env_name):
        return {}

    combined_dicts = merge_dicts(
        cast("dict[Any, Any]", env_dicts.get("*", {})),
        cast("dict[Any, Any]", env_dicts.get(env_name, {})),
    )
    return flatten_path_lists(combined_dicts, env_root)


def find_cfn_output(key: str, outputs: list[OutputTypeDef]) -> str | None:
    """Return CFN output value.

    Args:
        key: Name of the output.
        outputs: List of Stack outputs.

    """
    for i in outputs:
        if i.get("OutputKey") == key:
            return i.get("OutputValue")
    return None


def get_embedded_lib_path() -> str:
    """Return path of embedded libraries."""
    return str(Path(__file__).resolve().parent / "embedded")


def get_hash_for_filename(filename: str, hashfile_path: str) -> str:
    """Return hash for filename in the hashfile."""
    filehash = ""
    with open(hashfile_path, encoding="utf-8") as stream:  # noqa: PTH123
        for _cnt, line in enumerate(stream):
            if line.rstrip().endswith(filename):
                match = re.match(r"^[A-Za-z0-9]*", line)
                if match:
                    filehash = match.group(0)
                    break
    if filehash:
        return filehash
    raise AttributeError(f"Filename {filename} not found in hash file")


@contextmanager
def ignore_exit_code_0() -> Iterator[None]:
    """Capture exit calls and ignore those with exit code 0."""
    try:
        yield
    except SystemExit as exit_exc:
        if exit_exc.code != 0:
            raise


def fix_windows_command_list(commands: list[str]) -> list[str]:
    """Return command list with working Windows commands.

    npm on windows is npm.cmd, which will blow up
    subprocess.check_call(['npm', '...'])

    Similar issues arise when calling python apps that will have a windows-only
    suffix applied to them.

    """
    fully_qualified_cmd_path = which(commands[0])
    if fully_qualified_cmd_path:
        commands[0] = Path(fully_qualified_cmd_path).name
    return commands


def run_commands(
    commands: list[dict[str, list[str] | str] | list[str] | str],
    directory: Path | str,
    env: dict[str, str] | None = None,
) -> None:
    """Run list of commands."""
    directory = Path(directory)
    if env is None:
        env = os.environ.copy()
    for step in commands:
        if isinstance(step, (list, str)):
            execution_dir = directory
            raw_command = step
        elif step.get("command"):  # dictionary
            execution_dir = (
                directory / str(step["cwd"]) if step.get("cwd") and step["cwd"] else directory
            )
            raw_command = step["command"]
        else:
            raise AttributeError(f"Invalid command step: {step}")
        command_list = raw_command.split(" ") if isinstance(raw_command, str) else raw_command
        if platform.system().lower() == "windows":
            command_list = fix_windows_command_list(command_list)

        with change_dir(execution_dir):
            failed_to_find_error = (
                f'Attempted to run "{command_list[0]}" and failed to find it (are you sure it is '
                "installed and added to your PATH?)"
            )
            try:
                check_call(command_list, env=env)
            except FileNotFoundError:
                print(failed_to_find_error, file=sys.stderr)  # noqa: T201
                sys.exit(1)


def get_file_hash(
    filename: str,
    algorithm: Literal[
        "blake2b",
        "blake2b",
        "md5",
        "sha1",
        "sha224",
        "sha256",
        "sha3_224",
        "sha3_256",
        "sha3_384",
        "sha3_512",
        "sha384",
        "sha512",
        "shake_128",
        "shake_256",
    ],
) -> str:
    """Return cryptographic hash of file.

    .. deprecated:: 2.4.0
        Use :class:`runway.utils.FileHash` instead.

    """
    LOGGER.warning(
        "%s.get_file_hash is deprecated and will be removed in the next major release",
        __name__,
    )
    file_hash = getattr(hashlib, algorithm)()
    with open(filename, "rb") as stream:  # noqa: PTH123
        while True:
            data = stream.read(65536)  # 64kb chunks
            if not data:
                break
            file_hash.update(data)
    return file_hash.hexdigest()


def md5sum(filename: str) -> str:
    """Return MD5 hash of file.

    .. deprecated:: 2.4.0
        Use :class:`runway.utils.FileHash` instead.

    """
    LOGGER.warning(
        "%s.md5sum is deprecated and will be removed in the next major release",
        __name__,
    )
    return get_file_hash(filename, "md5")


def sha256sum(filename: str) -> str:
    """Return SHA256 hash of file.

    .. deprecated:: 2.4.0
        Use :class:`runway.utils.FileHash` instead.

    """
    LOGGER.warning(
        "%s.sha256sum is deprecated and will be removed in the next major release",
        __name__,
    )
    sha256 = hashlib.sha256()
    with open(filename, "rb", buffering=0) as stream:  # noqa: PTH123
        mem_view = memoryview(bytearray(128 * 1024))
        for i in iter(lambda: stream.readinto(mem_view), 0):
            sha256.update(mem_view[:i])
    return sha256.hexdigest()


@contextmanager
def use_embedded_pkgs(embedded_lib_path: str | None = None) -> Iterator[None]:
    """Temporarily prepend embedded packages to sys.path."""
    if embedded_lib_path is None:
        embedded_lib_path = get_embedded_lib_path()

    old_sys_path = list(sys.path)
    sys.path.insert(1, embedded_lib_path)  # https://stackoverflow.com/a/10097543
    try:
        yield
    finally:
        sys.path = old_sys_path


def which(program: str) -> str | None:
    """Mimic 'which' command behavior."""

    def is_exe(fpath: str) -> bool:
        """Determine if program exists and is executable."""
        return Path(fpath).is_file() and os.access(fpath, os.X_OK)

    def get_extensions() -> list[str]:
        """Get PATHEXT if the exist, otherwise use default."""
        exts = ".COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC"

        if os.getenv("PATHEXT"):
            exts = os.environ["PATHEXT"]

        return exts.split(";")

    fname, file_ext = os.path.splitext(program)  # noqa: PTH122
    fpath, fname = os.path.split(program)

    if not file_ext and platform.system().lower() == "windows":
        fnames = [fname + ext for ext in get_extensions()]
    else:
        fnames = [fname]

    for i in fnames:
        if fpath:
            exe_file = os.path.join(fpath, i)  # noqa: PTH118
            if is_exe(exe_file):
                return exe_file
        else:
            for path in (
                os.environ.get("PATH", "").split(os.pathsep)
                if "PATH" in os.environ
                else [os.getcwd()]  # noqa: PTH109
            ):
                exe_file = os.path.join(path, i)  # noqa: PTH118
                if is_exe(exe_file):
                    return exe_file

    return None



"""Runway Kubernetes Module options."""

from __future__ import annotations

from pathlib import Path

from pydantic import ConfigDict

from ...base import ConfigProperty


class RunwayK8sModuleOptionsDataModel(ConfigProperty):
    """Model for Runway Kubernetes Module options."""

    model_config = ConfigDict(
        extra="ignore",
        title="Runway Kubernetes Module options",
        validate_default=True,
        validate_assignment=True,
    )

    kubectl_version: str | None = None
    overlay_path: Path | None = None



"""Kubectl version management."""

from __future__ import annotations

import hashlib
import locale
import logging
import os
import platform
import re
import shutil
import sys
import tempfile
from typing import TYPE_CHECKING, Final, cast
from urllib.error import URLError
from urllib.request import urlretrieve

import requests

from ..compat import cached_property
from ..exceptions import KubectlVersionNotSpecified
from ..utils import FileHash, Version
from . import EnvManager, handle_bin_download_error

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path

    from .._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))
KB_VERSION_FILENAME = ".kubectl-version"
RELEASE_URI = "https://storage.googleapis.com/kubernetes-release/release"


def verify_kb_release(kb_url: str, download_dir: str, filename: str) -> None:
    """Compare checksum and exit if it doesn't match.

    Different releases provide varying checksum files. To account for this,
    start at SHA512 and work down to the first available checksum.

    requests is used for downloading these small files because of difficulty in
    getting 404 status from urllib on py2. Once py2 support is dropped, downloads
    can be moved to urllib.

    https://stackoverflow.com/questions/1308542/how-to-catch-404-error-in-urllib-urlretrieve

    """
    # This might be a bit cleaner refactored as self-referencing function, but
    # the ridiculousness should be short-lived as md5 & sha1 support won't last
    # long.
    try:
        hash_alg: hashlib._Hash = hashlib.sha512()
        checksum_filename = filename + "." + hash_alg.name
        LOGGER.debug("attempting download of kubectl %s checksum...", hash_alg.name)
        download_request = requests.get(
            kb_url + "/" + checksum_filename, allow_redirects=True, timeout=30
        )
        download_request.raise_for_status()
    except requests.exceptions.HTTPError:
        try:
            hash_alg = hashlib.sha256()
            checksum_filename = filename + "." + hash_alg.name
            LOGGER.debug("attempting download of kubectl %s checksum...", hash_alg.name)
            download_request = requests.get(
                kb_url + "/" + checksum_filename, allow_redirects=True, timeout=30
            )
            download_request.raise_for_status()
        except requests.exceptions.HTTPError:
            try:
                hash_alg = hashlib.sha1()  # noqa: S324
                checksum_filename = filename + "." + hash_alg.name
                LOGGER.debug("attempting download of kubectl %s checksum...", hash_alg.name)
                download_request = requests.get(
                    kb_url + "/" + checksum_filename, allow_redirects=True, timeout=30
                )
                download_request.raise_for_status()
            except requests.exceptions.HTTPError:
                try:
                    hash_alg = hashlib.md5()  # noqa: S324
                    checksum_filename = filename + "." + hash_alg.name
                    LOGGER.debug("attempting download of kubectl %s checksum...", hash_alg.name)
                    download_request = requests.get(
                        kb_url + "/" + checksum_filename, allow_redirects=True, timeout=30
                    )
                    download_request.raise_for_status()
                except requests.exceptions.HTTPError:
                    LOGGER.error("Unable to retrieve kubectl checksum file")
                    sys.exit(1)

    kb_hash = download_request.content.decode().rstrip("\n")

    checksum = FileHash(hash_alg)
    checksum.add_file(os.path.join(download_dir, filename))  # noqa: PTH118
    if kb_hash != checksum.hexdigest:
        LOGGER.error(
            "downloaded kubectl %s does not match %s checksum %s",
            filename,
            hash_alg.name,
            kb_hash,
        )
        sys.exit(1)
    LOGGER.debug("kubectl matched %s checksum...", hash_alg.name)


def download_kb_release(
    version: str,
    versions_dir: Path,
    kb_platform: str | None = None,
    arch: str | None = None,
) -> None:
    """Download kubectl and return path to it."""
    version_dir = versions_dir / version

    if arch is None:
        arch = os.getenv("KBENV_ARCH", "amd64")

    if not kb_platform:
        if platform.system().startswith("Darwin"):
            kb_platform = "darwin"
        elif platform.system().startswith("Windows") or (
            platform.system().startswith("MINGW64")
            or (
                platform.system().startswith("MSYS_NT")
                or (platform.system().startswith("CYGWIN_NT"))
            )
        ):
            kb_platform = "windows"
        else:
            kb_platform = "linux"

    download_dir = tempfile.mkdtemp()
    filename = "kubectl.exe" if kb_platform == "windows" else "kubectl"
    kb_url = f"{RELEASE_URI}/{version}/bin/{kb_platform}/{arch}"

    try:
        LOGGER.verbose("downloading kubectl from %s...", kb_url)
        urlretrieve(  # noqa: S310
            kb_url + "/" + filename, os.path.join(download_dir, filename)  # noqa: PTH118
        )
    except URLError as exc:
        handle_bin_download_error(exc, "kubectl")

    verify_kb_release(kb_url, download_dir, filename)

    version_dir.mkdir(parents=True, exist_ok=True)
    shutil.move(os.path.join(download_dir, filename), version_dir / filename)  # noqa: PTH118
    shutil.rmtree(download_dir)
    result = version_dir / filename
    result.chmod(result.stat().st_mode | 0o0111)  # ensure it is executable


class KBEnvManager(EnvManager):
    """kubectl version management.

    Designed to be compatible with https://github.com/alexppg/kbenv.

    """

    VERSION_REGEX: Final[str] = r"^(v)?(?P<version>[0-9]+\.[0-9]+\.[0-9]+\S*)"

    def __init__(self, path: Path | None = None, *, overlay_path: Path | None = None) -> None:
        """Initialize class.

        Args:
            path: Module path.
            overlay_path: Path to Kustomize overlay.

        """
        super().__init__("kubectl", "kbenv", path)
        self.overlay_path = overlay_path

    @cached_property
    def version(self) -> Version | None:
        """Terraform version."""
        if not self.current_version:
            self.current_version = self.get_version_from_file()
        if not self.current_version:
            return None
        return self.parse_version_string(self.current_version)

    @cached_property
    def version_file(self) -> Path | None:
        """Find and return a ".kubectl-version" file if one is present.

        Returns:
            Path to the kubectl version file.

        """
        path_list = [self.path, self.path.parent]
        if self.overlay_path:
            path_list.insert(0, self.overlay_path)
        for path in path_list:
            tmp_path = path / KB_VERSION_FILENAME
            if tmp_path.is_file():
                LOGGER.debug("using version file: %s", tmp_path)
                return tmp_path
        return None

    def get_version_from_file(self, file_path: Path | None = None) -> str | None:
        """Get kubectl version from a file.

        Args:
            file_path: Path to file that will be read.

        """
        file_path = file_path or self.version_file
        if file_path and file_path.is_file():
            return file_path.read_text(
                encoding=locale.getpreferredencoding(do_setlocale=False)
            ).strip()
        LOGGER.debug("file path not provided and version file could not be found")
        return None

    def install(self, version_requested: str | None = None) -> str:
        """Ensure kubectl is available."""
        if not version_requested:
            if self.version:
                version_requested = str(self.version)
            else:
                LOGGER.warning(
                    "kubectl version not specified and %s file not found",
                    KB_VERSION_FILENAME,
                )
                raise KubectlVersionNotSpecified

        if not version_requested.startswith("v"):
            version_requested = "v" + version_requested

        # Return early (i.e before reaching out to the internet) if the
        # matching version is already installed
        if (self.versions_dir / version_requested).is_dir():
            LOGGER.verbose("kubectl version %s already installed; using it...", version_requested)
            self.current_version = version_requested
            return str(self.bin)

        LOGGER.info("downloading and using kubectl version %s ...", version_requested)
        download_kb_release(version_requested, self.versions_dir)
        LOGGER.verbose("downloaded kubectl %s successfully", version_requested)
        self.current_version = version_requested
        return str(self.bin)

    def list_installed(self) -> Generator[Path, None, None]:
        """List installed versions of kubectl.

        Only lists versions of kubectl that have been installed by an instance
        if this class or by kbenv.

        """
        LOGGER.verbose("checking %s for kubectl versions...", self.versions_dir)
        return self.versions_dir.rglob("v*.*.*")

    def set_version(self, version: str) -> None:
        """Set current version.

        Clears cached values as needed.

        Args:
            version: Version string. Must be in the format of
                ``v<major>.<minor>.<patch>`` with an optional ``-<prerelease>``.

        """
        if self.current_version == version:
            return
        self.current_version = version
        self._del_cached_property("version")

    @classmethod
    def parse_version_string(cls, version: str) -> Version:
        """Parse version string into a :class:`Version`.

        Args:
            version: Version string to parse. Must be in the format of
                ``<major>.<minor>.<patch>`` with an optional ``-<prerelease>``.

        """
        match = re.search(cls.VERSION_REGEX, version)
        if not match:
            raise ValueError(f"provided version doesn't conform to regex: {cls.VERSION_REGEX}")
        return Version(f"v{match.group('version')}")



"""Retrieve a value from an environment variable."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from .base import LookupHandler

if TYPE_CHECKING:

    from ...context import CfnginContext, RunwayContext


class EnvLookup(LookupHandler["CfnginContext | RunwayContext"]):
    """Environment variable Lookup."""

    TYPE_NAME: ClassVar[str] = "env"
    """Name that the Lookup is registered as."""

    @classmethod
    def handle(cls, value: str, context: CfnginContext | RunwayContext, **_kwargs: Any) -> Any:
        """Retrieve an environment variable.

        The value is retrieved from a copy of the current environment variables
        that is saved to the context object. These environment variables
        are manipulated at runtime by Runway to fill in additional values
        such as ``DEPLOY_ENVIRONMENT`` and ``AWS_REGION`` to match the
        current execution.

        Args:
            value: The value passed to the Lookup.
            context: The current context object.

        Raises:
            ValueError: Unable to find a value for the provided query and
                a default value was not provided.

        """
        query, args = cls.parse(value)

        result = context.env.vars.get(query, args.pop("default", ""))

        if result != "":  # allows for False bool and NoneType results
            return cls.format_results(result, **args)

        raise ValueError(f'"{value}" does not exist in the environment')



"""AWS IAM hook."""

from __future__ import annotations

import copy
import logging
from typing import TYPE_CHECKING, Any, cast

from awacs import ecs
from awacs.aws import Allow, Policy, Statement
from awacs.helpers.trust import get_ecs_assumerole_policy
from botocore.exceptions import ClientError

from ...utils import BaseModel
from . import utils

if TYPE_CHECKING:
    from mypy_boto3_iam.type_defs import (
        GetServerCertificateResponseTypeDef,
        UploadServerCertificateResponseTypeDef,
    )

    from ...context import CfnginContext

LOGGER = logging.getLogger(__name__)

ECS_SERVICE_ROLE_NAME = "ecsServiceRole"
ECS_SERVICE_ROLE_POLICY = Policy(
    Version="2012-10-17",
    Statement=[
        Statement(
            Effect=Allow,
            Resource=["*"],
            Action=[
                ecs.CreateCluster,
                ecs.DeregisterContainerInstance,
                ecs.DiscoverPollEndpoint,
                ecs.Poll,
                ecs.Action("Submit*"),
            ],
        )
    ],
)


class CreateEcsServiceRoleHookArgs(BaseModel):
    """Hook arguments for ``create_ecs_service_role``."""

    role_name: str = ECS_SERVICE_ROLE_NAME
    """Name of the role to create."""


class EnsureServerCertExistsHookArgs(BaseModel):
    """Hook arguments for ``ensure_server_cert_exists``."""

    cert_name: str
    """Name of the certificate that should exist."""

    path_to_certificate: str | None = None
    """Path to certificate file."""

    path_to_chain: str | None = None
    """Path to chain file."""

    path_to_private_key: str | None = None
    """Path to private key file."""

    prompt: bool = True
    """Whether to prompt to upload a certificate if one does not exist."""


def create_ecs_service_role(context: CfnginContext, *_args: Any, **kwargs: Any) -> bool:
    """Create ecsServiceRole IAM role.

    https://docs.aws.amazon.com/AmazonECS/latest/developerguide/using-service-linked-roles.html

    Args:
        context: Context instance. (passed in by CFNgin)
        **kwargs: Arbitrary keyword arguments.

    """
    args = CreateEcsServiceRoleHookArgs.model_validate(kwargs)
    client = context.get_session().client("iam")

    try:
        client.create_role(
            RoleName=args.role_name,
            AssumeRolePolicyDocument=get_ecs_assumerole_policy().to_json(),
        )
    except ClientError as err:
        if "already exists" not in str(err):
            raise
    client.put_role_policy(
        RoleName=args.role_name,
        PolicyName="AmazonEC2ContainerServiceRolePolicy",
        PolicyDocument=ECS_SERVICE_ROLE_POLICY.to_json(),
    )
    return True


def _get_cert_arn_from_response(
    response: GetServerCertificateResponseTypeDef | UploadServerCertificateResponseTypeDef,
) -> str:
    result = copy.deepcopy(response)
    # GET response returns this extra key
    if "ServerCertificate" in response:
        return cast("GetServerCertificateResponseTypeDef", result)["ServerCertificate"][
            "ServerCertificateMetadata"
        ]["Arn"]
    return (
        cast("UploadServerCertificateResponseTypeDef", result)
        .get("ServerCertificateMetadata", {"Arn": ""})
        .get("Arn", "")
    )


def _get_cert_contents(kwargs: dict[str, Any]) -> dict[str, Any]:  # noqa: C901
    """Build parameters with server cert file contents.

    Args:
        kwargs: The keyword args passed to ensure_server_cert_exists, optionally
            containing the paths to the cert, key and chain files.

    Returns:
        A dictionary containing the appropriate parameters to supply to
        upload_server_certificate. An empty dictionary if there is a problem.

    """
    paths = {
        "certificate": kwargs.get("path_to_certificate"),
        "private_key": kwargs.get("path_to_private_key"),
        "chain": kwargs.get("path_to_chain"),
    }

    for key, value in paths.items():
        if value is not None:
            continue

        path = input(f"Path to {key} (skip): ")
        if path == "skip" or not path.strip():
            continue

        paths[key] = path

    parameters: dict[str, str] = {}

    for key, path in paths.items():
        if not path:
            continue

        # Allow passing of file like object for tests
        try:
            contents = path.read()
        except AttributeError:
            with open(utils.full_path(path), encoding="utf-8") as read_file:  # noqa: PTH123
                contents = read_file.read()

        if key == "certificate":
            parameters["CertificateBody"] = contents
        elif key == "private_key":
            parameters["PrivateKey"] = contents
        elif key == "chain":
            parameters["CertificateChain"] = contents

    if parameters and "cert_name" in kwargs:
        parameters["ServerCertificateName"] = kwargs["cert_name"]

    return parameters


def ensure_server_cert_exists(context: CfnginContext, *_args: Any, **kwargs: Any) -> dict[str, str]:
    """Ensure server cert exists.

    Args:
        context: CFNgin context object.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        Dict containing ``status``, ``cert_name``, and ``cert_arn``.

    """
    args = EnsureServerCertExistsHookArgs.model_validate(kwargs)
    client = context.get_session().client("iam")
    status = "unknown"
    try:
        response = client.get_server_certificate(ServerCertificateName=args.cert_name)
        cert_arn = _get_cert_arn_from_response(response)
        status = "exists"
        LOGGER.info("certificate exists: %s (%s)", args.cert_name, cert_arn)
    except ClientError:
        if args.prompt:
            upload = input(f"Certificate '{args.cert_name}' wasn't found. Upload it now? (yes/no) ")
            if upload != "yes":
                return {}

        parameters = _get_cert_contents(args.model_dump())
        if not parameters:
            return {}
        response = client.upload_server_certificate(**parameters)
        cert_arn = _get_cert_arn_from_response(response)
        status = "uploaded"
        LOGGER.info("uploaded certificate: %s (%s)", args.cert_name, cert_arn)

    return {
        "status": status,
        "cert_name": args.cert_name,
        "cert_arn": cert_arn,
    }



"""Retrieve a value from SSM Parameter Store."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar

from ...lookups.handlers.base import LookupHandler

if TYPE_CHECKING:
    from mypy_boto3_ssm.type_defs import ParameterTypeDef

    from ...context import CfnginContext, RunwayContext

LOGGER = logging.getLogger(__name__)


class SsmLookup(LookupHandler["CfnginContext | RunwayContext"]):
    """SSM Parameter Store Lookup."""

    TYPE_NAME: ClassVar[str] = "ssm"
    """Name that the Lookup is registered as."""

    @classmethod
    def handle(cls, value: str, context: CfnginContext | RunwayContext, **_kwargs: Any) -> Any:
        """Retrieve a value from SSM Parameter Store.

        Args:
            value: The value passed to the Lookup.
            context: The current context object.

        Raises:
            ParameterNotFound: Parameter not found in SSM and a default value
                was not provided.

        """
        query, args = cls.parse(value)

        session = context.get_session(region=args.get("region"))
        client = session.client("ssm")

        try:
            return cls.format_results(
                cls._handle_get_parameter(
                    client.get_parameter(Name=query, WithDecryption=True)["Parameter"]
                ),
                **args,
            )
        except client.exceptions.ParameterNotFound:
            if args.get("default"):
                args.pop("load", None)  # don't load a default value
                return cls.format_results(args.pop("default"), **args)
            raise

    @staticmethod
    def _handle_get_parameter(parameter: ParameterTypeDef) -> list[str] | str | None:
        """Handle the return value of ``get_parameter``."""
        if "Value" not in parameter:
            return None
        value = parameter["Value"]
        if parameter.get("Type") == "StringList":
            return value.split(",")
        return value



#!/usr/bin/env python
"""Module with Terraform state resources."""
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

import awacs.dynamodb
import awacs.s3
from awacs.aws import Allow, PolicyDocument, Statement
from troposphere import Equals, If, Join, NoValue, Or, Output, dynamodb, iam, s3

from runway.cfngin.blueprints.base import Blueprint
from runway.cfngin.blueprints.variables.types import CFNString

if TYPE_CHECKING:
    from runway.cfngin.blueprints.type_defs import BlueprintVariableTypeDef


class TfState(Blueprint):
    """CFNgin blueprint for creating Terraform state resources."""

    VARIABLES: ClassVar[dict[str, BlueprintVariableTypeDef]] = {
        "BucketDeletionPolicy": {
            "type": str,
            "allowed_values": ["Delete", "Retain"],
            "description": "CloudFormation deletion policy",
            "default": "Retain",
        },
        "BucketName": {
            "type": CFNString,
            "description": "(optional) Name for the S3 bucket",
            "default": "",
        },
        "TableName": {
            "type": CFNString,
            "description": "(optional) Name for the DynamoDB table",
            "default": "",
        },
    }

    def create_template(self) -> None:
        """Create template (main function called by CFNgin)."""
        self.template.set_version("2010-09-09")
        self.template.set_description("Terraform State Resources")

        # Conditions
        for i in ["BucketName", "TableName"]:
            self.template.add_condition(
                f"{i}Omitted",
                Or(
                    Equals(self.variables[i].ref, ""),
                    Equals(self.variables[i].ref, "undefined"),
                ),
            )

        # Resources
        terraformlocktable = self.template.add_resource(
            dynamodb.Table(
                "TerraformStateTable",
                AttributeDefinitions=[
                    dynamodb.AttributeDefinition(AttributeName="LockID", AttributeType="S")
                ],
                KeySchema=[dynamodb.KeySchema(AttributeName="LockID", KeyType="HASH")],
                ProvisionedThroughput=dynamodb.ProvisionedThroughput(
                    ReadCapacityUnits=2, WriteCapacityUnits=2
                ),
                TableName=If("TableNameOmitted", NoValue, self.variables["TableName"].ref),
            )
        )
        self.template.add_output(
            Output(
                f"{terraformlocktable.title}Name",
                Description="Name of DynamoDB table for Terraform state",
                Value=terraformlocktable.ref(),
            )
        )

        terraformstatebucket = self.template.add_resource(
            s3.Bucket(
                "TerraformStateBucket",
                DeletionPolicy=self.variables["BucketDeletionPolicy"],
                AccessControl=s3.Private,
                BucketName=If("BucketNameOmitted", NoValue, self.variables["BucketName"].ref),
                LifecycleConfiguration=s3.LifecycleConfiguration(
                    Rules=[s3.LifecycleRule(NoncurrentVersionExpirationInDays=90, Status="Enabled")]
                ),
                VersioningConfiguration=s3.VersioningConfiguration(Status="Enabled"),
            )
        )
        self.template.add_output(
            Output(
                f"{terraformstatebucket.title}Name",
                Description="Name of bucket storing Terraform state",
                Value=terraformstatebucket.ref(),
            )
        )
        self.template.add_output(
            Output(
                f"{terraformstatebucket.title}Arn",
                Description="Arn of bucket storing Terraform state",
                Value=terraformstatebucket.get_att("Arn"),
            )
        )

        managementpolicy = self.template.add_resource(
            iam.ManagedPolicy(
                "ManagementPolicy",
                Description="Managed policy for Terraform state management.",
                Path="/",
                PolicyDocument=PolicyDocument(
                    Version="2012-10-17",
                    Statement=[
                        # https://www.terraform.io/docs/backends/types/s3.html#s3-bucket-permissions
                        Statement(
                            Action=[awacs.s3.ListBucket],
                            Effect=Allow,
                            Resource=[terraformstatebucket.get_att("Arn")],
                        ),
                        Statement(
                            Action=[awacs.s3.GetObject, awacs.s3.PutObject],
                            Effect=Allow,
                            Resource=[Join("", [terraformstatebucket.get_att("Arn"), "/*"])],
                        ),
                        Statement(
                            Action=[
                                awacs.dynamodb.GetItem,
                                awacs.dynamodb.PutItem,
                                awacs.dynamodb.DeleteItem,
                            ],
                            Effect=Allow,
                            Resource=[terraformlocktable.get_att("Arn")],
                        ),
                    ],
                ),
            )
        )
        self.template.add_output(
            Output(
                "PolicyArn",
                Description="Managed policy Arn",
                Value=managementpolicy.ref(),
            )
        )


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    from runway.context import CfnginContext

    print(TfState("test", CfnginContext(parameters={"namespace": "test"})).to_json())  # noqa: T201



"""Runway Static Site Module options."""

from ._components import StaticSiteOptions
from ._models import (
    RunwayStaticSiteExtraFileDataModel,
    RunwayStaticSiteModuleOptionsDataModel,
    RunwayStaticSitePreBuildStepDataModel,
    RunwayStaticSiteSourceHashingDataModel,
    RunwayStaticSiteSourceHashingDirectoryDataModel,
)

__all__ = [
    "RunwayStaticSiteExtraFileDataModel",
    "RunwayStaticSiteModuleOptionsDataModel",
    "RunwayStaticSitePreBuildStepDataModel",
    "RunwayStaticSiteSourceHashingDataModel",
    "RunwayStaticSiteSourceHashingDirectoryDataModel",
    "StaticSiteOptions",
]



"""Runway config base definition."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from ...._logging import PrefixAdaptor
from ....exceptions import UnresolvedVariable
from ....variables import Variable

if TYPE_CHECKING:
    from typing_extensions import Self

    from ...._logging import RunwayLogger
    from ....context import RunwayContext
    from ...models.base import ConfigProperty
    from ._variables_def import RunwayVariablesDefinition

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))

_ConfigPropertyTypeVar = TypeVar("_ConfigPropertyTypeVar", bound="ConfigProperty")


class ConfigComponentDefinition(ABC, Generic[_ConfigPropertyTypeVar]):
    """Base class for Runway config components."""

    _data: _ConfigPropertyTypeVar
    _pre_process_vars: tuple[str, ...] = ()
    _supports_vars: tuple[str, ...] = ()
    _vars: dict[str, Variable] = {}

    def __init__(self, data: _ConfigPropertyTypeVar) -> None:
        """Instantiate class."""
        self._data = data.model_copy(deep=True)

        self._vars = {}
        for var in self._supports_vars:
            if self._data[var]:
                self._register_variable(var, self._data[var])

    @property
    def data(self) -> dict[str, Any]:
        """Return the underlying data as a dict."""
        return self._data.model_dump()

    def get(self, name: str, default: Any = None) -> None:
        """Get a value or return default if it is not found.

        Args:
            name: The value to look for.
            default: Returned if no other value is found.

        """
        return getattr(self, name, default)

    def resolve(
        self,
        context: RunwayContext,
        *,
        pre_process: bool = False,
        variables: RunwayVariablesDefinition | None = None,
    ) -> None:
        """Resolve variables.

        Args:
            context: Runway context object.
            pre_process: Whether to only resolve pre-process fields.
            variables: Object containing values to resolve the ``var`` lookup.

        """
        logger = PrefixAdaptor(self.name, LOGGER) if hasattr(self, "name") else LOGGER

        if pre_process:
            logger.verbose("resolving variables for pre-processing...")
            for field in self._pre_process_vars:
                if field in self._vars:
                    self._vars[field].resolve(context, variables=variables)
                    self._data[field] = self._vars[field].value
            return

        logger.verbose("resolving variables...")
        for field, var in self._vars.items():
            var.resolve(context, variables=variables)
            self._data[field] = var.value

    def _register_variable(self, var_name: str, var_value: Any) -> None:
        """Register a variable.

        Called while instantiating the class, this is responsible for initializing
        fields that support variables.

        It can be overridden by subclasses to alter how the variable are initialized.

        Args:
            var_name: Name of the config field that can contain a variable
                lookup.
            var_value: Literal value supplied in the config to be resolved
                as a variable if it contains a lookup.

        """
        self._vars[var_name] = Variable(name=var_name, value=var_value, variable_type="runway")

    @classmethod
    @abstractmethod
    def parse_obj(cls: type[Self], obj: object) -> Self:
        """Parse a python object into this class.

        Args:
            obj: The object to parse.

        """
        raise NotImplementedError

    def __contains__(self, name: str) -> bool:
        """Implement evaluation of 'in' conditional."""
        if name.startswith("_"):
            return name in self.__dict__
        return self._data.__contains__(name)

    def __getattr__(self, name: str) -> Any:
        """Implement evaluation of self.name.

        Args:
            name: The value to look for.

        Raises:
            AttributeError: Object does not contain an attribute for the
                name provided.
            UnresolvedVariable: The value being access is a variable and it
                has not been resolved yet.

        """
        if name in self._vars and not self._vars[name].resolved:
            raise UnresolvedVariable(self._vars[name])
        if name in super().__getattribute__("_data"):
            return super().__getattribute__("_data").__getitem__(name)
        raise AttributeError(f"{self.__class__.__name__} object has no attribute {name}")

    def __getitem__(self, name: str) -> Any:
        """Implement evaluation of self[name].

        Args:
            name: The value to look for.

        Raises:
            KeyError: Object does not contain a field of the name provided.

        """
        try:
            return self.__getattr__(name)
        except AttributeError as exc:
            raise KeyError(name) from exc

    def __setattr__(self, name: str, value: Any) -> None:
        """Implement evaluation of self.name = value.

        When setting an attribute, the value is set on the underlying data model.
        The exception to this is if the name starts with an underscore.

        Args:
            name: The value to set.
            value: The value to assigned to the field.

        Raises:
            AttributeError: The name being set is a property without a setter.

        """
        prop = getattr(self.__class__, name, None)
        if isinstance(prop, property) and prop.fset:
            prop.fset(self, value)
        elif isinstance(prop, property):
            raise AttributeError(f"setting {name} property is not supported")
        elif name.startswith("_") or name in dir(self):
            super().__setattr__(name, value)
        else:
            self._data[name] = value

    def __setitem__(self, name: str, value: Any) -> None:
        """Implement evaluation of self[name] = value.

        When setting an attribute, the value is set on the underlying data model.
        The exception to this is if the name starts with an underscore.

        Args:
            name: The value to set.
            value: The value to assigned to the field.

        """
        self.__setattr__(name, value)



"""Base sync strategies.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/syncstrategy/base.py

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar

from typing_extensions import Literal, Self

if TYPE_CHECKING:
    from botocore.session import Session

    from ..file_generator import FileStats
    from ..parameters import ParametersDataModel


LOGGER = logging.getLogger(__name__.replace("._", "."))

ValidSyncType = Literal["file_at_src_and_dest", "file_not_at_dest", "file_not_at_src"]
VALID_SYNC_TYPES: list[ValidSyncType] = [
    "file_at_src_and_dest",
    "file_not_at_dest",
    "file_not_at_src",
]


class BaseSync:
    """Base sync strategy.

    To create a new sync strategy, subclass from this class.

    """

    NAME: ClassVar = None

    sync_type: ValidSyncType

    def __init__(self, sync_type: ValidSyncType = "file_at_src_and_dest") -> None:
        """Instantiate class.

        Args:
            sync_type: This determines where the sync strategy will be
                used. There are three strings to choose from.

        """
        self._check_sync_type(sync_type)
        self.sync_type = sync_type

    @property
    def name(self) -> str | None:
        """Retrieve the ``name`` of the sync strategy's ``ARGUMENT``."""
        return self.NAME

    @staticmethod
    def _check_sync_type(sync_type: str) -> None:
        if sync_type not in VALID_SYNC_TYPES:
            raise ValueError(
                f"Unknown sync_type: {sync_type}.\nValid options are {VALID_SYNC_TYPES}."
            )

    def register_strategy(self, session: Session) -> None:
        """Register the sync strategy class to the given session."""
        session.register("choosing-s3-sync-strategy", self.use_sync_strategy)

    def determine_should_sync(
        self, src_file: FileStats | None, dest_file: FileStats | None
    ) -> bool:
        """Determine if file should sync.

        This function takes two ``FileStat`` objects (one from the source and
        one from the destination). Then makes a decision on whether a given
        operation (e.g. a upload, copy, download) should be allowed
        to take place.

        The function currently raises a ``NotImplementedError``. So this
        method must be overwritten when this class is subclassed. Note
        that this method must return a Boolean as documented below.

        Args:
            src_file: A representation of the operation that is to be
                performed on a specific file existing in the source. Note if
                the file does not exist at the source, ``src_file`` is None.
            dest_file: A representation of the operation that is to be
                performed on a specific file existing in the destination. Note if
                the file does not exist at the destination, ``dest_file`` is None.

        Return:
            True if an operation based on the ``FileStat`` should be allowed to occur.
            False if if an operation based on the ``FileStat`` should not be
            allowed to occur. Note the operation being referred to depends on
            the ``sync_type`` of the sync strategy:

            'file_at_src_and_dest': refers to ``src_file``

            'file_not_at_dest': refers to ``src_file``

            'file_not_at_src': refers to ``dest_file``

        """
        raise NotImplementedError("determine_should_sync")

    def use_sync_strategy(self, params: ParametersDataModel, **_: Any) -> Self | None:
        """Determine which sync strategy to use.

        The sync strategy object must be returned by this method
        if it is to be chosen as the sync strategy to use.

        Args:
            params: All arguments that a sync strategy is able to process.

        """
        if self.name and params.get(self.name):
            # Return the sync strategy object to be used for syncing.
            return self
        return None

    @staticmethod
    def compare_size(src_file: FileStats | None, dest_file: FileStats | None) -> bool:
        """Compare the size of two FileStats objects."""
        if not (src_file and dest_file):
            raise ValueError("src_file and dest_file must not be None")
        return src_file.size == dest_file.size

    def compare_time(self, src_file: FileStats | None, dest_file: FileStats | None) -> bool:
        """Compare modified time of two FileStats objects.

        Returns:
            True if the file does not need updating based on time of last
            modification and type of operation. False if the file does need
            updating based on the time of last modification and type of operation.

        """
        if not (src_file and dest_file):
            raise ValueError("src_file and dest_file must not be None")
        delta = dest_file.last_update - src_file.last_update
        cmd = src_file.operation_name
        if cmd in ["copy", "upload"]:
            return delta.total_seconds() >= 0
        return bool(cmd == "download" and delta.total_seconds() <= 0)


class MissingFileSync(BaseSync):
    """File is missing from destination."""

    def __init__(self, sync_type: Literal["file_not_at_dest"] = "file_not_at_dest") -> None:
        """Instantiate class.

        Args:
            sync_type: This determines where the sync strategy will be
                used. There are three strings to choose from.

        """
        super().__init__(sync_type)

    def determine_should_sync(
        self, src_file: FileStats | None, dest_file: FileStats | None  # noqa: ARG002
    ) -> bool:
        """Determine if file should sync."""
        LOGGER.debug(
            "syncing: %s -> %s, file does not exist at destination",
            src_file.src if src_file else None,
            src_file.dest if src_file else None,
        )
        return True


class NeverSync(BaseSync):
    """Never sync file."""

    def __init__(self, sync_type: Literal["file_not_at_src"] = "file_not_at_src") -> None:
        """Instantiate class.

        Args:
            sync_type: This determines where the sync strategy will be
                used. There are three strings to choose from.

        """
        super().__init__(sync_type)

    def determine_should_sync(
        self, src_file: FileStats | None, dest_file: FileStats | None  # noqa: ARG002
    ) -> bool:
        """Determine if file should sync."""
        return False


class SizeAndLastModifiedSync(BaseSync):
    """Sync based on size and last modified date."""

    def determine_should_sync(
        self, src_file: FileStats | None, dest_file: FileStats | None
    ) -> bool:
        """Determine if file should sync."""
        same_size = self.compare_size(src_file, dest_file)
        same_last_modified_time = self.compare_time(src_file, dest_file)
        should_sync = (not same_size) or (not same_last_modified_time)
        if should_sync:
            LOGGER.debug(
                "syncing: %s -> %s, size: %s -> %s, modified time: %s -> %s",
                src_file.src if src_file else None,
                src_file.dest if src_file else None,
                src_file.size if src_file else None,
                dest_file.size if dest_file else None,
                src_file.last_update if src_file else None,
                dest_file.last_update if dest_file else None,
            )
        return should_sync



"""Runway static site Module parameters."""

from __future__ import annotations

from pydantic import ConfigDict, Field, field_validator

from ....config.models.base import ConfigProperty


class RunwayStaticSiteCustomErrorResponseDataModel(ConfigProperty):
    """Model for Runway stat site Module staticsite_custom_error_responses parameter item.

    https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-cloudfront-distribution-customerrorresponse.html

    """

    model_config = ConfigDict(
        extra="forbid",
        title="Runway static site Module staticsite_custom_error_responses parameter item",
        validate_default=True,
        validate_assignment=True,
    )

    ErrorCachingMinTTL: int | None = None
    ErrorCode: int | None = None
    ResponseCode: int | None = None
    ResponsePagePath: str | None = None


class RunwayStaticSiteLambdaFunctionAssociationDataModel(ConfigProperty):
    """Model for Runway stat site Module staticsite_lambda_function_associations parameter item."""

    model_config = ConfigDict(
        extra="forbid",
        title="Runway static site Module staticsite_lambda_function_associations parameter item",
        validate_default=True,
        validate_assignment=True,
    )

    arn: str
    """Lambda function ARN."""

    type: str
    """Association type."""


def _staticsite_alias_generator(field_name: str) -> str:
    """Append ``staticsite`` to field names.

    Some fields are excluded from having aliases (e.g. namespace).

    """
    return f"staticsite_{field_name}" if field_name != "namespace" else field_name


class RunwayStaticSiteModuleParametersDataModel(ConfigProperty):
    """Model for Runway static site Module parameters."""

    model_config = ConfigDict(
        alias_generator=_staticsite_alias_generator,
        extra="ignore",
        populate_by_name=True,
        title="Runway static site Module parameters",
        validate_default=True,
        validate_assignment=True,
    )

    acmcert_arn: str | None = None
    """The certificate arn used for any alias domains supplied.
    This is a requirement when supplying any custom domain.

    """

    additional_redirect_domains: list[str] = []
    """Additional domains (beyond the ``aliases`` domains or the CloudFront URL if
    no aliases are provided) that will be authorized by the Auth@Edge UserPool AppClient.

    """

    aliases: list[str] = []
    """Any custom domains that should be added to the CloudFront Distribution."""

    auth_at_edge: bool = False
    """Auth@Edge make the static site private by placing it behind an authorization wall."""

    cf_disable: bool = False
    """Whether deployment of the CloudFront Distribution should be disabled."""

    compress: bool = True
    """Whether the CloudFront default cache behavior will automatically compress certain files."""

    cookie_settings: dict[str, str] = {
        "idToken": "Path=/; Secure; SameSite=Lax",
        "accessToken": "Path=/; Secure; SameSite=Lax",
        "refreshToken": "Path=/; Secure; SameSite=Lax",
        "nonce": "Path=/; Secure; HttpOnly; Max-Age=1800; SameSite=Lax",
    }
    """The default cookie settings for retrieved tokens and generated nonce's."""

    create_user_pool: bool = False
    """Whether to create a User Pool for the Auth@Edge configuration."""

    custom_error_responses: list[RunwayStaticSiteCustomErrorResponseDataModel] = []
    """Define custom error responses."""

    enable_cf_logging: bool = True
    """Enable CloudFront logging."""

    http_headers: dict[str, str] = {
        "Content-Security-Policy": "default-src https: 'unsafe-eval' 'unsafe-inline'; "
        "font-src 'self' 'unsafe-inline' 'unsafe-eval' data: https:; "
        "object-src 'none'; "
        "connect-src 'self' https://*.amazonaws.com https://*.amazoncognito.com",
        "Strict-Transport-Security": "max-age=31536000; includeSubdomains; preload",
        "Referrer-Policy": "same-origin",
        "X-XSS-Protection": "1; mode=block",
        "X-Frame-Options": "DENY",
        "X-Content-Type-Options": "nosniff",
    }
    """Headers that should be sent with each origin response."""

    lambda_function_associations: list[RunwayStaticSiteLambdaFunctionAssociationDataModel] = []
    """This allows the user to deploy custom Lambda@Edge associations with their pre-build function versions."""

    namespace: str
    """The unique namespace for the deployment."""

    non_spa: bool = False
    """Whether this site is a single page application (SPA)."""

    oauth_scopes: list[str] = [
        "phone",
        "email",
        "profile",
        "openid",
        "aws.cognito.signin.user.admin",
    ]
    """Scope is a mechanism in OAuth 2.0 to limit an application's access to a user's account."""

    redirect_path_auth_refresh: str = "/refreshauth"
    """The path that a user is redirected to when their authorization tokens have expired (1 hour)."""

    redirect_path_sign_in: str = "/parseauth"
    """The path that a user is redirected to after sign-in."""

    redirect_path_sign_out: str = "/"
    """The path that a user is redirected to after sign-out."""

    required_group: str | None = None
    """Name of Cognito User Pool group of which users must be a member to be granted access to the site.
    If ``None``, allows all UserPool users to have access.

    """

    rewrite_directory_index: str | None = None
    """Deploy a Lambda@Edge function designed to rewrite directory indexes."""

    role_boundary_arn: str | None = None
    """Defines an IAM Managed Policy that will be set as the permissions boundary
    for any IAM Roles created to support the site.

    """

    service_role: str | None = Field(default=None, alias="cloudformation_service_role")
    """IAM role that CloudFormation will use."""

    sign_out_url: str = "/signout"
    """The path a user should access to sign themselves out of the application."""

    supported_identity_providers: list[str] = ["COGNITO"]
    """A comma delimited list of the User Pool client identity providers."""

    user_pool_arn: str | None = None
    """The ARN of a pre-existing Cognito User Pool to use with Auth@Edge."""

    web_acl: str | None = None
    """The ARN of a web access control list (web ACL) to associate with the CloudFront Distribution."""

    @field_validator(
        "additional_redirect_domains",
        "aliases",
        "supported_identity_providers",
        mode="before",
    )
    @classmethod
    def _convert_comma_delimited_list(cls, v: list[str] | str) -> list[str]:
        """Convert comma delimited lists to a string."""
        if isinstance(v, str):
            return [i.strip() for i in v.split(",")]
        return v



"""DynamoDB lookup."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, ClassVar, cast

from botocore.exceptions import ClientError
from typing_extensions import Literal, TypedDict

from ....lookups.handlers.base import LookupHandler
from ....utils import BaseModel
from ...utils import read_value_from_path

if TYPE_CHECKING:
    from mypy_boto3_dynamodb.type_defs import AttributeValueTypeDef

    from ....context import CfnginContext
    from ....lookups.handlers.base import ParsedArgsTypeDef


_QUERY_PATTERN = r"""(?x)  # <table_name>@<partition_key>:<partition_key_value>.<attribute>
^(?P<table_name>[a-zA-Z0-9\-_\.]{3,255})  # name of the DynamoDB Table
@  # delimiter
(?P<partition_key>\S*)  # partition/primary key
:  # delimiter
(?P<partition_key_value>[^\.]*)  # value of partition/primary key
\.  # delimiter
(?P<attribute>.*)$  # attribute to get
"""
"""Lookup query pattern minus region argument.

.. note::
    This pattern and/or it's variable will likely change in a future release so it
    should not be consumed directly by any external code.

"""


class ArgsDataModel(BaseModel):
    """Arguments data model."""

    region: str | None = None
    """AWS region."""


class QueryDataModel(BaseModel):
    """Arguments data model."""

    attribute: str
    """The attribute to be returned by this lookup.
    Supports additional syntax to retrieve a nested value.

    """

    partition_key: str
    """The DynamoDB Table's partition key."""

    partition_key_value: str
    """The value of the partition key to query the database."""

    table_name: str
    """Name of the DynamoDB Table to query."""

    @property
    def item_key(self) -> dict[str, AttributeValueTypeDef]:
        """Value to pass to boto3 ``.get_item()`` call as the ``Key`` argument.

        Raises:
            ValueError: The value of ``partition_key_value`` doesn't match the
                required regex and so it can't be parsed.

        """
        pattern = re.compile(r"^(?P<value>[^\[]+)\[?(?P<data_type>[BNS]+)?]?$")
        match = pattern.search(self.partition_key_value)
        if not match:
            raise ValueError(
                f"Partition key value '{self.partition_key_value}' "
                f"doesn't match regex: {pattern.pattern}"
            )
        return {
            self.partition_key: cast(
                "AttributeValueTypeDef",
                {match.groupdict("S")["data_type"]: match.group("value")},
            )
        }


class DynamodbLookup(LookupHandler["CfnginContext"]):
    """DynamoDB lookup."""

    TYPE_NAME: ClassVar[str] = "dynamodb"
    """Name that the Lookup is registered as."""

    @classmethod
    def parse(cls, value: str) -> tuple[str, ParsedArgsTypeDef]:
        """Parse the value passed to the lookup.

        This overrides the default parsing to account for special requirements.

        Args:
            value: The raw value passed to a lookup.

        Returns:
            The lookup query and a dict of arguments

        Raises:
            ValueError: The value provided does not appear to contain the name of
                a DynamoDB Table. The name of a Table is required.

        """
        raw_value = read_value_from_path(value)
        args: ParsedArgsTypeDef = {}

        if "@" not in raw_value:
            raise ValueError(
                f"'{raw_value}' missing delimiter for DynamoDB Table name:\n{_QUERY_PATTERN}"
            )

        table_info, table_keys = raw_value.split("@", 1)
        if ":" in table_info:
            args["region"], table_info = table_info.split(":", 1)

        return f"{table_info}@{table_keys}", args

    @classmethod
    def parse_query(cls, value: str) -> QueryDataModel:
        """Parse query string to extract. Does not support arguments in ``value``.

        Raises:
            ValueError: The argument provided does not match the expected format defined
                with a regex pattern.

        """
        # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/HowItWorks.NamingRulesDataTypes.html
        pattern = re.compile(_QUERY_PATTERN)
        match = pattern.search(value)
        if not match:
            raise ValueError(f"Query '{value}' doesn't match regex:\n{pattern.pattern}")
        return QueryDataModel.model_validate(match.groupdict())

    @classmethod
    def handle(cls, value: str, context: CfnginContext, *__args: Any, **__kwargs: Any) -> Any:
        """Get a value from a DynamoDB table.

        Args:
            value: Parameter(s) given to this lookup.
                ``[<region>:]<tablename>@<partitionkey>:<keyvalue>.<keyvalue>...``
            context: Context instance.

        Raises:
            ValueError: The value provided to the lookup resulted in an error.

        .. note:: The region is optional, and defaults to the environment's
                  ``AWS_DEFAULT_REGION`` if not specified.

        """
        raw_query, raw_args = cls.parse(value)
        query = cls.parse_query(raw_query)
        args = ArgsDataModel.model_validate(raw_args)

        table_keys = query.attribute.split(".")

        key_dict = _lookup_key_parse(table_keys)

        dynamodb = context.get_session(region=args.region).client("dynamodb")
        try:
            response = dynamodb.get_item(
                TableName=query.table_name,
                Key=query.item_key,
                ProjectionExpression=",".join([query.partition_key, *key_dict["clean_table_keys"]]),
            )
        except dynamodb.exceptions.ResourceNotFoundException as exc:
            raise ValueError(f"Can't find the DynamoDB table: {query.table_name}") from exc
        except ClientError as exc:
            if exc.response["Error"]["Code"] == "ValidationException":
                raise ValueError(
                    f"No DynamoDB record matched the partition key: {query.partition_key}"
                ) from exc
            raise ValueError(f"The DynamoDB lookup '{value}' encountered an error: {exc}") from exc
        # find and return the key from the dynamo data returned
        if "Item" in response:
            return _get_val_from_ddb_data(response["Item"], key_dict["new_keys"])
        raise ValueError(
            f"The DynamoDB record could not be found using the following: {query.item_key}"
        )


class ParsedLookupKey(TypedDict):
    """Return value of _lookup_key_parse."""

    clean_table_keys: list[str]
    new_keys: list[dict[Literal["L", "M", "N", "S"], str]]


def _lookup_key_parse(table_keys: list[str]) -> ParsedLookupKey:
    """Return the order in which the stacks should be executed.

    Args:
        table_keys: List of keys a table.

    Returns:
        Includes a dict of lookup types with data types ('new_keys')
        and a list of the lookups without ('clean_table_keys')

    Raises:
        ValueError: DynamoDB data type is not supported.

    """
    # we need to parse the key lookup passed in
    regex_matcher = r"\[([^\]]+)]"
    valid_dynamodb_datatypes = ["L", "M", "N", "S"]
    clean_table_keys: list[str] = []
    new_keys: list[dict[Literal["L", "M", "N", "S"], str]] = []

    for key in table_keys:
        match = re.search(regex_matcher, key)
        if match:
            # the datatypes are pulled from the dynamodb docs
            if match.group(1) not in valid_dynamodb_datatypes:
                raise ValueError(
                    f"CFNgin does not support looking up the data type: {match.group(1)}"
                )
            match_val = cast(Literal["L", "M", "N", "S"], match.group(1))
            key = key.replace(match.group(0), "")  # noqa: PLW2901
            new_keys.append({match_val: key})
        else:
            new_keys.append({"S": key})
        clean_table_keys.append(key)
    return {"new_keys": new_keys, "clean_table_keys": clean_table_keys}


def _get_val_from_ddb_data(
    data: dict[str, Any], keylist: list[dict[Literal["L", "M", "N", "S"], str]]
) -> Any:
    """Return the value of the lookup.

    Args:
        data: The raw DynamoDB data.
        keylist: A list of keys to lookup. This must include the datatype.

    Returns:
        The value from the DynamoDB record, and casts it to a matching python
        datatype.

    """
    next_type: str | None = None
    # iterate through the keylist to find the matching key/datatype
    for key in keylist:
        for k in key:
            if next_type is None:
                data = data[key[k]]
            else:
                temp_dict = data[next_type]
                data = temp_dict[key[k]]
            next_type = k
    if next_type == "L":
        # if type is list, convert it to a list and return
        return _convert_ddb_list_to_list(data[cast(str, next_type)])
    if next_type == "N":
        # TODO (troyready): handle various types of 'number' datatypes, (e.g. int, double)
        # if a number, convert to an int and return
        return int(data[cast(str, next_type)])
    # else, just assume its a string and return
    return str(data[cast(str, next_type)])


def _convert_ddb_list_to_list(conversion_list: list[dict[str, Any]]) -> list[Any]:
    """Return a python list without the DynamoDB datatypes.

    Args:
        conversion_list: A DynamoDB list which includes the datatypes.

    Returns:
        Returns A sanitized list without the datatypes.

    """
    return [val[v] for val in conversion_list for v in val]



"""Runway variables."""

from __future__ import annotations

import logging
import re
from collections.abc import Iterable, Iterator, MutableMapping, MutableSequence
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast, overload

from pydantic import BaseModel
from typing_extensions import Literal

from .cfngin.lookups.registry import CFNGIN_LOOKUP_HANDLERS
from .exceptions import (
    FailedLookup,
    FailedVariableLookup,
    InvalidLookupConcatenation,
    UnknownLookupType,
    UnresolvedVariable,
    UnresolvedVariableValue,
)
from .lookups.registry import RUNWAY_LOOKUP_HANDLERS

if TYPE_CHECKING:
    from .cfngin.providers.aws.default import Provider
    from .config.components.runway import RunwayVariablesDefinition
    from .context import CfnginContext, RunwayContext
    from .lookups.handlers.base import LookupHandler

LOGGER = logging.getLogger(__name__)

_LiteralValue = TypeVar("_LiteralValue", int, str)
_PydanticModelTypeVar = TypeVar("_PydanticModelTypeVar", bound=BaseModel)
VariableTypeLiteralTypeDef = Literal["cfngin", "runway"]


class Variable:
    """Represents a variable provided to a Runway directive."""

    _value: VariableValue
    name: str

    def __init__(
        self,
        name: str,
        value: Any,
        variable_type: VariableTypeLiteralTypeDef = "cfngin",
    ) -> None:
        """Initialize class.

        Args:
            name: Name of the variable (directive/key).
            value: The variable itself.
            variable_type: Type of variable (cfngin|runway).

        """
        self.name = name
        self._raw_value = value
        self._value = VariableValue.parse_obj(value, variable_type)
        self.variable_type = variable_type

    @property
    def dependencies(self) -> set[str]:
        """Stack names that this variable depends on.

        Returns:
            set[str]: Stack names that this variable depends on.

        """
        return self._value.dependencies

    @property
    def resolved(self) -> bool:
        """Boolean for whether the Variable has been resolved.

        Variables only need to be resolved if they contain lookups.

        """
        return self._value.resolved

    @property
    def value(self) -> Any:
        """Return the current value of the Variable.

        Raises:
            UnresolvedVariable: Value accessed before it have been resolved.

        """
        try:
            return self._value.value
        except UnresolvedVariableValue:
            raise UnresolvedVariable(self) from None

    def resolve(
        self,
        context: CfnginContext | RunwayContext,
        provider: Provider | None = None,
        variables: RunwayVariablesDefinition | None = None,
        **kwargs: Any,
    ) -> None:
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            FailedVariableLookup

        """
        try:
            self._value.resolve(context, provider=provider, variables=variables, **kwargs)
        except FailedLookup as err:
            raise FailedVariableLookup(self, err) from err.cause

    def get(self, key: str, default: Any = None) -> Any:
        """Implement evaluation of self.get.

        Args:
            key: Attribute name to return the value for.
            default: Value to return if attribute is not found.

        """
        return getattr(self.value, key, default)

    def __repr__(self) -> str:
        """Return object representation."""
        return f"Variable[{self.name}={self._raw_value}]"


def resolve_variables(
    variables: list[Variable],
    context: CfnginContext | RunwayContext,
    provider: Provider | None = None,
) -> None:
    """Given a list of variables, resolve all of them.

    Args:
        variables: List of variables.
        context: CFNgin context.
        provider: Subclass of the base provider.

    """
    for variable in variables:
        variable.resolve(context=context, provider=provider)


_VariableValue = TypeVar("_VariableValue", bound="VariableValue")


class VariableValue:
    """Syntax tree base class to parse variable values."""

    _resolved: bool = False
    _data: Any
    variable_type: VariableTypeLiteralTypeDef

    @property
    def dependencies(self) -> set[Any]:
        """Stack names that this variable depends on."""
        return set()

    @property
    def resolved(self) -> bool:
        """Use to check if the variable value has been resolved.

        Raises:
            NotImplementedError: Should be defined in a subclass.

        """
        raise NotImplementedError

    @property
    def simplified(self) -> Any:
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or
        flatten nested concatenations.

        Should be implimented in subclasses where applicable.

        """
        return self

    @property
    def value(self) -> Any:
        """Value of the variable. Can be resolved or unresolved.

        Raises:
            NotImplementedError: Should be defined in a subclass.

        """
        raise NotImplementedError

    def resolve(
        self,
        context: CfnginContext | RunwayContext,
        provider: Provider | None = None,
        variables: RunwayVariablesDefinition | None = None,
        **kwargs: Any,
    ) -> None:
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.
            **kwargs: Arbitrary keyword arguments.

        """

    def _resolve(self, value: Any) -> None:
        """Set _value and _resolved from the result of resolve().

        Args:
            value: Resolved value of the variable.

        """
        self._data = value
        self._resolved = True

    @overload
    @classmethod
    def parse_obj(
        cls, obj: _PydanticModelTypeVar, variable_type: VariableTypeLiteralTypeDef = ...
    ) -> VariableValuePydanticModel[_PydanticModelTypeVar]: ...

    @overload
    @classmethod
    def parse_obj(
        cls, obj: dict[str, Any], variable_type: VariableTypeLiteralTypeDef = ...
    ) -> VariableValue: ...

    @overload
    @classmethod
    def parse_obj(
        cls, obj: list[Any], variable_type: VariableTypeLiteralTypeDef = ...
    ) -> VariableValueList: ...

    @overload
    @classmethod
    def parse_obj(
        cls, obj: int, variable_type: VariableTypeLiteralTypeDef = ...
    ) -> VariableValueLiteral[int]: ...

    @overload
    @classmethod
    def parse_obj(
        cls, obj: str, variable_type: VariableTypeLiteralTypeDef = ...
    ) -> VariableValueConcatenation[VariableValueLiteral[str] | VariableValueLookup]: ...

    @classmethod
    def parse_obj(  # noqa: C901
        cls, obj: Any, variable_type: VariableTypeLiteralTypeDef = "cfngin"
    ) -> VariableValue:
        """Parse complex variable structures using type appropriate subclasses.

        Args:
            obj: The objected defined as the value of a variable.
            variable_type: Type of variable (cfngin|runway).

        """
        if isinstance(obj, BaseModel):
            return VariableValuePydanticModel(obj, variable_type=variable_type)
        if isinstance(obj, dict):
            return VariableValueDict(obj, variable_type=variable_type)  # type: ignore
        if isinstance(obj, list):
            return VariableValueList(obj, variable_type=variable_type)  # type: ignore
        if not isinstance(obj, str):
            return VariableValueLiteral(obj, variable_type=variable_type)

        tokens: VariableValueConcatenation[VariableValueLiteral[str] | VariableValueLookup] = (
            VariableValueConcatenation(
                # pyright 1.1.138 is having issues properly inferring the type from comprehension
                [
                    VariableValueLiteral(cast(str, t), variable_type=variable_type)
                    for t in re.split(r"(\$\{|\}|\s+)", obj)  # ${ or space or }
                ]
            )
        )

        opener = "${"
        closer = "}"

        while True:
            last_open = None
            next_close = None
            for i, tok in enumerate(tokens):
                if not isinstance(tok, VariableValueLiteral):
                    continue
                if tok.value == opener:
                    last_open = i
                    next_close = None
                if last_open is not None and tok.value == closer and next_close is None:
                    next_close = i

            if next_close is not None:
                lookup_query = VariableValueConcatenation(
                    tokens[(cast(int, last_open) + len(opener) + 1) : next_close],
                    variable_type=variable_type,
                )
                lookup = VariableValueLookup(
                    lookup_name=tokens[cast(int, last_open) + 1],  # type: ignore
                    lookup_query=lookup_query,
                    variable_type=variable_type,
                )
                tokens[last_open : (next_close + 1)] = [lookup]  # type: ignore
            else:
                break  # cov: ignore

        return tokens.simplified

    def __iter__(self) -> Iterator[Any]:
        """How the object is iterated.

        Raises:
            NotImplementedError: Should be defined in a subclass.

        """
        raise NotImplementedError

    def __repr__(self) -> str:
        """Return object representation.

        Raises:
            NotImplementedError: Should be defined in a subclass.

        """
        raise NotImplementedError


class VariableValueDict(VariableValue, MutableMapping[str, VariableValue]):
    """A dict variable value."""

    def __init__(
        self, data: dict[str, Any], variable_type: VariableTypeLiteralTypeDef = "cfngin"
    ) -> None:
        """Instantiate class.

        Args:
            data: Data to be stored in the object.
            variable_type: Type of variable (cfngin|runway).

        """
        self._data = {k: self.parse_obj(v, variable_type=variable_type) for k, v in data.items()}
        self.variable_type: VariableTypeLiteralTypeDef = variable_type

    @property
    def dependencies(self) -> set[str]:
        """Stack names that this variable depends on."""
        deps: set[str] = set()
        for item in self.values():
            deps.update(item.dependencies)
        return deps

    @property
    def resolved(self) -> bool:
        """Use to check if the variable value has been resolved."""
        accumulator: bool = True
        for item in self.values():
            accumulator = accumulator and item.resolved
        return accumulator

    @property
    def simplified(self) -> dict[str, Any]:
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or
        flatten nested concatenations.

        """
        return {k: v.simplified for k, v in self.items()}

    @property
    def value(self) -> dict[str, Any]:
        """Value of the variable. Can be resolved or unresolved."""
        return {k: v.value for k, v in self.items()}

    def resolve(
        self,
        context: CfnginContext | RunwayContext,
        provider: Provider | None = None,
        variables: RunwayVariablesDefinition | None = None,
        **kwargs: Any,
    ) -> None:
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.
            **kwargs: Arbitrary keyword arguments.

        """
        for item in self.values():
            item.resolve(context, provider=provider, variables=variables, **kwargs)

    def __delitem__(self, __key: str) -> None:
        """Delete item by index."""
        del self._data[__key]

    def __getitem__(self, __key: str) -> VariableValue:
        """Get item by index."""
        return self._data[__key]

    def __iter__(self) -> Iterator[str]:
        """How the object is iterated."""
        yield from iter(self._data)

    def __len__(self) -> int:
        """Length of the object."""
        return len(self._data)

    def __repr__(self) -> str:
        """Return object representation."""
        return f"dict[{', '.join(f'{k}={v}' for k, v in self.items())}]"

    def __setitem__(self, __key: str, __value: VariableValue) -> None:
        """Set item by index."""
        self._data[__key] = __value


class VariableValueList(VariableValue, MutableSequence[VariableValue]):
    """List variable value."""

    def __init__(
        self,
        iterable: Iterable[Any],
        variable_type: VariableTypeLiteralTypeDef = "cfngin",
    ) -> None:
        """Instantiate class.

        Args:
            iterable: Data to store in the iterable.
            variable_type: Type of variable (cfngin|runway).

        """
        self._data: list[VariableValue] = [
            self.parse_obj(i, variable_type=variable_type) for i in iterable
        ]
        self.variable_type: VariableTypeLiteralTypeDef = variable_type

    @property
    def dependencies(self) -> set[str]:
        """Stack names that this variable depends on."""
        deps: set[str] = set()
        for item in self:
            deps.update(item.dependencies)
        return deps

    @property
    def resolved(self) -> bool:
        """Use to check if the variable value has been resolved."""
        accumulator: bool = True
        for item in self:
            accumulator = accumulator and item.resolved
        return accumulator

    @property
    def simplified(self) -> list[VariableValue]:
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or
        flatten nested concatenations.

        """
        return [item.simplified for item in self]

    @property
    def value(self) -> list[Any]:
        """Value of the variable. Can be resolved or unresolved."""
        return [item.value for item in self]

    def insert(self, index: int, value: VariableValue) -> None:
        """Insert a value at a specific index."""
        self._data.insert(index, value)

    def resolve(
        self,
        context: CfnginContext | RunwayContext,
        provider: Provider | None = None,
        variables: RunwayVariablesDefinition | None = None,
        **kwargs: Any,
    ) -> None:
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.
            **kwargs: Arbitrary keyword arguments.

        """
        for item in self:
            item.resolve(context, provider=provider, variables=variables, **kwargs)

    @overload
    def __delitem__(self, index: int) -> None: ...

    @overload
    def __delitem__(self, index: slice) -> None: ...

    def __delitem__(self, index: int | slice) -> None:
        """Delete item by index."""
        del self._data[index]

    @overload
    def __getitem__(self, __index: int) -> VariableValue: ...

    @overload
    def __getitem__(self, __index: slice) -> list[VariableValue]: ...

    def __getitem__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, __index: int | slice
    ) -> MutableSequence[VariableValue] | VariableValue:
        """Get item by index."""
        return self._data[__index]

    @overload
    def __setitem__(self, __index: int, __value: VariableValue) -> None: ...

    @overload
    def __setitem__(self, __index: slice, __value: list[VariableValue]) -> None: ...

    def __setitem__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        __index: int | slice,
        __value: list[VariableValue] | VariableValue,
    ) -> None:
        """Set item by index."""
        self._data[__index] = __value  # type: ignore

    def __iter__(self) -> Iterator[VariableValue]:
        """Object iteration."""
        yield from iter(self._data)

    def __len__(self) -> int:
        """Length of the object."""
        return len(self._data)

    def __repr__(self) -> str:
        """Object string representation."""
        return f"list[{', '.join(repr(i) for i in self._data)}]"


class VariableValueLiteral(Generic[_LiteralValue], VariableValue):
    """The literal value of a variable as provided."""

    def __init__(
        self, value: _LiteralValue, variable_type: VariableTypeLiteralTypeDef = "cfngin"
    ) -> None:
        """Instantiate class.

        Args:
            value: Data to store in the object.
            variable_type: Type of variable (cfngin|runway).

        """
        self._data = value
        self.variable_type: VariableTypeLiteralTypeDef = variable_type

    @property
    def resolved(self) -> bool:
        """Use to check if the variable value has been resolved.

        The ValueLiteral will always appear as resolved because it does
        not "resolve" since it is the literal definition of the value.

        """
        return True

    @property
    def value(self) -> _LiteralValue:
        """Value of the variable."""
        return self._data

    def __iter__(self) -> Iterator[Any]:
        """How the object is iterated."""
        yield self

    def __repr__(self) -> str:
        """Return object representation."""
        return f"Literal[{self._data}]"


class VariableValueConcatenation(Generic[_VariableValue], VariableValue):
    """A concatenated variable values."""

    def __init__(
        self,
        iterable: Iterable[_VariableValue],
        variable_type: VariableTypeLiteralTypeDef = "cfngin",
    ) -> None:
        """Instantiate class.

        Args:
            iterable: Data to store in the iterable.
            variable_type: Type of variable (cfngin|runway).

        """
        self._data = list(iterable)
        self.variable_type: VariableTypeLiteralTypeDef = variable_type

    @property
    def dependencies(self) -> set[str]:
        """Stack names that this variable depends on."""
        deps: set[str] = set()
        for item in self:
            deps.update(item.dependencies)
        return deps

    @property
    def resolved(self) -> bool:
        """Use to check if the variable value has been resolved."""
        accumulator: bool = True
        for item in self:
            accumulator = accumulator and item.resolved
        return accumulator

    @property
    def simplified(self) -> VariableValue:
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or flatten
        nested concatenations.

        """
        concat: list[VariableValue] = []
        for item in self:
            if isinstance(item, VariableValueLiteral) and item.value == "":
                pass
            elif (
                isinstance(item, VariableValueLiteral)
                and concat
                and isinstance(concat[-1], VariableValueLiteral)
            ):
                concat[-1] = VariableValueLiteral(
                    str(concat[-1].value) + str(item.value)  # type: ignore
                )
            elif isinstance(item, VariableValueConcatenation):
                concat.extend(iter(item.simplified))
            else:
                concat.append(item.simplified)

        if not concat:
            return VariableValueLiteral("")
        if len(concat) == 1:
            return concat[0]
        return VariableValueConcatenation(concat)

    @property
    def value(self) -> Any:
        """Value of the variable. Can be resolved or unresolved.

        Raises:
            InvalidLookupConcatenation

        """
        if len(self) == 1:
            return self[0].value

        values: list[str] = []
        for value in self:
            resolved_value = value.value
            if isinstance(resolved_value, bool) or not isinstance(resolved_value, (int, str)):
                raise InvalidLookupConcatenation(value, self)
            values.append(str(resolved_value))
        return "".join(values)

    def resolve(
        self,
        context: CfnginContext | RunwayContext,
        provider: Provider | None = None,
        variables: RunwayVariablesDefinition | None = None,
        **kwargs: Any,
    ) -> None:
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.
            **kwargs: Arbitrary keyword arguments.

        """
        for value in self:
            value.resolve(context, provider=provider, variables=variables, **kwargs)

    def __delitem__(self, __index: int) -> None:
        """Delete item by index."""
        del self._data[__index]

    @overload
    def __getitem__(self, __index: int) -> _VariableValue: ...

    @overload
    def __getitem__(self, __index: slice) -> list[_VariableValue]: ...

    def __getitem__(self, __index: int | slice) -> list[_VariableValue] | _VariableValue:
        """Get item by index."""
        return self._data[__index]

    @overload
    def __setitem__(self, __index: int, __value: _VariableValue) -> None: ...

    @overload
    def __setitem__(self, __index: slice, __value: list[_VariableValue]) -> None: ...

    def __setitem__(
        self,
        __index: int | slice,
        __value: list[_VariableValue] | _VariableValue,
    ) -> None:
        """Set item by index."""
        self._data[__index] = __value

    def __iter__(self) -> Iterator[_VariableValue]:
        """Object iteration."""
        yield from iter(self._data)

    def __len__(self) -> int:
        """Length of the object."""
        return len(self._data)

    def __repr__(self) -> str:
        """Return object representation."""
        return f"Concatenation[{', '.join(repr(v) for v in self)}]"


class VariableValueLookup(VariableValue):
    """A lookup variable value."""

    handler: type[LookupHandler[Any]]
    lookup_name: VariableValueLiteral[str]
    lookup_query: VariableValue

    _resolved: bool

    def __init__(
        self,
        lookup_name: VariableValueLiteral[str],
        lookup_query: str | VariableValue,
        handler: type[LookupHandler[Any]] | None = None,
        variable_type: VariableTypeLiteralTypeDef = "cfngin",
    ) -> None:
        """Initialize class.

        Args:
            lookup_name: Name of the invoked lookup.
            lookup_query: Data portion of the lookup.
            handler: Lookup handler that will be use to resolve the value.
            variable_type: Type of variable (cfngin|runway).

        Raises:
            UnknownLookupType: Invalid lookup type.
            ValueError: Invalid value for variable_type.

        """
        self._resolved = False
        self._data = None

        self.lookup_name = lookup_name
        self.variable_type: VariableTypeLiteralTypeDef = variable_type

        if isinstance(lookup_query, str):
            lookup_query = VariableValueLiteral(lookup_query)
        self.lookup_query = lookup_query

        if handler is None:
            lookup_name_resolved = lookup_name.value
            try:
                if variable_type == "cfngin":
                    handler = CFNGIN_LOOKUP_HANDLERS[lookup_name_resolved]
                elif variable_type == "runway":
                    handler = RUNWAY_LOOKUP_HANDLERS[lookup_name_resolved]
                else:
                    raise ValueError('Variable type must be one of "cfngin" or "runway"')
            except KeyError:
                raise UnknownLookupType(self) from None
        self.handler = handler

    @property
    def dependencies(self) -> set[str]:
        """Stack names that this variable depends on."""
        if hasattr(self.handler, "dependencies"):
            return self.handler.dependencies(self.lookup_query)
        return set()

    @property
    def resolved(self) -> bool:
        """Use to check if the variable value has been resolved."""
        return self._resolved

    @property
    def simplified(self) -> VariableValueLookup:
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or
        flatten nested concatenations.

        """
        return self

    @property
    def value(self) -> Any:
        """Value of the variable. Can be resolved or unresolved.

        Raises:
            UnresolvedVariableValue: Value accessed before it has been resolved.

        """
        if self._resolved:
            return self._data
        raise UnresolvedVariableValue(self)

    def resolve(
        self,
        context: CfnginContext | RunwayContext,
        provider: Provider | None = None,
        variables: RunwayVariablesDefinition | None = None,
        **kwargs: Any,
    ) -> None:
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            FailedLookup: A lookup failed for any reason.

        """
        self.lookup_query.resolve(context=context, provider=provider, variables=variables, **kwargs)
        try:
            result = self.handler.handle(
                self.lookup_query.value,
                context=context,
                provider=provider,
                variables=variables,
                **kwargs,
            )
            return self._resolve(result)
        except Exception as err:
            raise FailedLookup(self, err) from err

    def __iter__(self) -> Iterator[VariableValueLookup]:
        """How the object is iterated."""
        yield self

    def __repr__(self) -> str:
        """Return object representation."""
        if self._resolved:
            return f"Lookup[{self._data} ({self.lookup_name} {self.lookup_query!r})]"
        return f"Lookup[{self.lookup_name} {self.lookup_query!r}]"

    def __str__(self) -> str:
        """Object displayed as a string."""
        return f"${{{self.lookup_name.value} {self.lookup_query.value}}}"


class VariableValuePydanticModel(Generic[_PydanticModelTypeVar], VariableValue):
    """A pydantic model variable value."""

    def __init__(
        self,
        data: _PydanticModelTypeVar,
        variable_type: VariableTypeLiteralTypeDef = "cfngin",
    ) -> None:
        """Instantiate class.

        Args:
            data: Data to be stored in the object.
            variable_type: Type of variable (cfngin|runway).

        """
        self._data: dict[str, VariableValue] = {
            k: self.parse_obj(v, variable_type=variable_type) for k, v in data
        }
        self._model_class = type(data)
        self.variable_type: VariableTypeLiteralTypeDef = variable_type

    @property
    def dependencies(self) -> set[str]:
        """Stack names that this variable depends on."""
        deps: set[str] = set()
        for value in self._data.values():
            deps.update(value.dependencies)
        return deps

    @property
    def resolved(self) -> bool:
        """Use to check if the variable value has been resolved."""
        accumulator: bool = True
        for value in self._data.values():
            accumulator = accumulator and value.resolved
        return accumulator

    @property
    def simplified(self) -> dict[str, Any]:
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or
        flatten nested concatenations.

        """
        return {field: value.simplified for field, value in self._data.items()}

    @property
    def value(self) -> _PydanticModelTypeVar:
        """Value of the variable. Can be resolved or unresolved.

        Uses the original pydantic model class to parse the resolved data back
        into a pydantic model.

        """
        return self._model_class.model_validate(
            {field: value.value for field, value in self._data.items()}
        )

    def resolve(
        self,
        context: CfnginContext | RunwayContext,
        provider: Provider | None = None,
        variables: RunwayVariablesDefinition | None = None,
        **kwargs: Any,
    ) -> None:
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.
            **kwargs: Arbitrary keyword arguments.

        """
        for item in self._data.values():
            item.resolve(context, provider=provider, variables=variables, **kwargs)

    def __delitem__(self, __key: str) -> None:
        """Delete item by index."""
        del self._data[__key]

    def __getitem__(self, __key: str) -> VariableValue:
        """Get item by index."""
        return self._data[__key]

    def __iter__(self) -> Iterator[str]:
        """How the object is iterated."""
        yield from iter(self._data)

    def __len__(self) -> int:
        """Length of the object."""
        return len(self._data)

    def __repr__(self) -> str:
        """Return object representation."""
        return (
            self._model_class.__name__ + f"[{', '.join(f'{k}={v}' for k, v in self._data.items())}]"
        )

    def __setitem__(self, __key: str, __value: VariableValue) -> None:
        """Set item by index."""
        self._data[__key] = __value



"""Type definitions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Union

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

RunwayEnvironmentsType: TypeAlias = dict[str, Union[bool, list[str], str]]
RunwayEnvironmentsUnresolvedType: TypeAlias = Union[dict[str, Union[bool, list[str], str]], str]
RunwayEnvVarsType: TypeAlias = dict[str, Union[list[str], str]]
RunwayEnvVarsUnresolvedType: TypeAlias = Union[RunwayEnvVarsType, str]
RunwayModuleTypeTypeDef: TypeAlias = Literal[
    "cdk", "cloudformation", "kubernetes", "serverless", "static", "terraform"
]



"""CFNgin config models."""

from __future__ import annotations

import copy
import locale
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, cast

import yaml
from pydantic import ConfigDict, Field, field_validator, model_validator
from typing_extensions import Literal

from .. import utils
from ..base import ConfigProperty
from ._package_sources import (
    CfnginPackageSourcesDefinitionModel,
    GitCfnginPackageSourceDefinitionModel,
    LocalCfnginPackageSourceDefinitionModel,
    S3CfnginPackageSourceDefinitionModel,
)

if TYPE_CHECKING:
    from pydantic.config import JsonDict
    from typing_extensions import Self

__all__ = [
    "CfnginConfigDefinitionModel",
    "CfnginHookDefinitionModel",
    "CfnginPackageSourcesDefinitionModel",
    "CfnginStackDefinitionModel",
    "GitCfnginPackageSourceDefinitionModel",
    "LocalCfnginPackageSourceDefinitionModel",
    "S3CfnginPackageSourceDefinitionModel",
]


class CfnginHookDefinitionModel(ConfigProperty):
    """Model for a CFNgin hook definition."""

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "description": "Python classes or functions run before or after deploy/destroy actions."
        },
        title="CFNgin Hook Definition",
        validate_default=True,
        validate_assignment=True,
    )

    args: Annotated[
        dict[str, Any],
        Field(
            title="Arguments",
            description="Arguments that will be passed to the hook. (supports lookups)",
        ),
    ] = {}
    data_key: Annotated[
        str | None, Field(description="Key to use when storing the returned result of the hook.")
    ] = None
    enabled: Annotated[bool, Field(description="Whether the hook will be run.")] = True
    path: Annotated[str, Field(description="Python importable path to the hook.")]
    required: Annotated[
        bool, Field(description="Whether to continue execution if the hook results in an error.")
    ] = True


@staticmethod
def _stack_json_schema_extra(schema: JsonDict) -> None:
    """Process the schema after it has been generated.

    Schema is modified in place. Return value is ignored.

    https://pydantic-docs.helpmanual.io/usage/schema/#schema-customization

    """
    schema["description"] = "Define CloudFormation stacks using a Blueprint or Template."

    # prevents a false error when defining stacks as a dict
    if "required" in schema and isinstance(schema["required"], list):
        schema["required"].remove("name")

    # fields that can be bool or lookup
    if "properties" in schema and isinstance(schema["properties"], dict):
        properties = schema["properties"]
        for field_name in ["enabled", "locked", "protected", "termination_protection"]:
            if field_name in properties and isinstance(properties[field_name], dict):
                field_schema = cast("JsonDict", properties[field_name])
                field_schema.pop("type")
                field_schema["anyOf"] = [
                    {"type": "boolean"},
                    {"type": "string", "pattern": utils.CFNGIN_LOOKUP_STRING_REGEX},
                ]


class CfnginStackDefinitionModel(ConfigProperty):
    """Model for a CFNgin stack definition."""

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra=_stack_json_schema_extra,
        title="CFNgin Stack Definition",
        validate_default=True,
        validate_assignment=True,
    )

    class_path: Annotated[
        str | None,
        Field(
            title="Blueprint Class Path", description="Python importable path to a blueprint class."
        ),
    ] = None
    """Python importable path to a blueprint class."""

    description: Annotated[
        str | None,
        Field(
            title="Stack Description",
            description="A description that will be applied to the stack in CloudFormation.",
        ),
    ] = None
    """A description that will be applied to the stack in CloudFormation."""

    enabled: Annotated[bool, Field(description="Whether the stack will be deployed.")] = True
    """Whether the stack will be deployed."""

    in_progress_behavior: Annotated[
        Literal["wait"] | None,
        Field(
            title="Stack In Progress Behavior",
            description="The action to take when a stack's status is "
            "CREATE_IN_PROGRESS or UPDATE_IN_PROGRESS when trying to update it.",
        ),
    ] = None
    """The action to take when a Stack's status is ``CREATE_IN_PROGRESS`` or
    ``UPDATE_IN_PROGRESS`` when trying to update it.

    """

    locked: Annotated[bool, Field(description="Whether to limit updating of the stack.")] = False
    """Whether to limit updating of the stack."""

    name: Annotated[str, Field(title="Stack Name", description="Name of the stack.")]
    """Name of the stack."""

    protected: Annotated[
        bool,
        Field(
            description="Whether to force all updates to the stack to be performed interactively."
        ),
    ] = False
    """Whether to force all updates to the stack to be performed interactively."""

    required_by: Annotated[
        list[str], Field(description="Array of stacks (by name) that require this stack.")
    ] = []
    """Array of stacks (by name) that require this stack."""

    requires: Annotated[
        list[str], Field(description="Array of stacks (by name) that this stack requires.")
    ] = []
    """Array of stacks (by name) that this stack requires."""

    stack_name: Annotated[
        str | None,
        Field(
            title="Explicit Stack Name",
            description="Explicit name of the stack (namespace will still be prepended).",
        ),
    ] = None
    """Explicit name of the stack (namespace will still be prepended)."""

    stack_policy_path: Annotated[
        Path | None,
        Field(
            description="Path to a stack policy document that will be applied to the CloudFormation stack."
        ),
    ] = None
    """Path to a stack policy document that will be applied to the CloudFormation stack."""

    tags: Annotated[
        dict[str, Any], Field(description="Tags that will be applied to the CloudFormation stack.")
    ] = {}
    """Tags that will be applied to the CloudFormation stack."""

    template_path: Annotated[
        Path | None, Field(description="Path to a JSON or YAML formatted CloudFormation Template.")
    ] = None
    """Path to a JSON or YAML formatted CloudFormation Template."""

    termination_protection: Annotated[
        bool,
        Field(description="Set the value of termination protection on the CloudFormation stack."),
    ] = False
    """Set the value of termination protection on the CloudFormation stack."""

    timeout: Annotated[
        int | None,
        Field(
            description="The amount of time (in minutes) that can pass before the Stack status becomes CREATE_FAILED."
        ),
    ] = None
    """The amount of time (in minutes) that can pass before the Stack status becomes CREATE_FAILED."""

    variables: Annotated[
        dict[str, Any],
        Field(
            description="Parameter values that will be passed to the Blueprint/CloudFormation stack. (supports lookups)"
        ),
    ] = {}
    """Parameter values that will be passed to the Blueprint/CloudFormation stack. (supports lookups)"""

    _resolve_path_fields = field_validator("stack_policy_path", "template_path")(
        utils.resolve_path_field
    )

    @model_validator(mode="before")
    @classmethod
    def _validate_class_and_template(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate class_path and template_path are not both provided."""
        if values.get("class_path") and values.get("template_path"):
            raise ValueError("only one of class_path or template_path can be defined")
        return values

    @model_validator(mode="before")
    @classmethod
    def _validate_class_or_template(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Ensure that either class_path or template_path is defined."""
        # if the Stack is disabled or locked, it is ok that these are missing
        required = values.get("enabled", True) and not values.get("locked", False)
        if not values.get("class_path") and not values.get("template_path") and required:
            raise ValueError("either class_path or template_path must be defined")
        return values


class CfnginConfigDefinitionModel(ConfigProperty):
    """Model for a CFNgin config definition."""

    model_config = ConfigDict(
        extra="ignore",
        json_schema_extra={"description": "Configuration file for Runway's CFNgin."},
        title="CFNgin Config File",
        validate_default=True,
        validate_assignment=True,
    )

    cfngin_bucket: Annotated[
        str | None,
        Field(
            title="CFNgin Bucket",
            description="Name of an AWS S3 bucket to use for caching CloudFormation templates. "
            "Set as an empty string to disable caching.",
        ),
    ] = None
    cfngin_bucket_region: Annotated[
        str | None,
        Field(
            title="CFNgin Bucket Region",
            description="AWS Region where the CFNgin Bucket is located. "
            "If not provided, the current region is used.",
        ),
    ] = None
    cfngin_cache_dir: Annotated[
        Path | None,
        Field(
            title="CFNgin Cache Directory",
            description="Path to a local directory that CFNgin will use for local caching.",
        ),
    ] = None
    log_formats: Annotated[  # TODO (kyle): create model
        dict[str, str], Field(description="Customize log message formatting by log level.")
    ] = {}
    lookups: Annotated[
        dict[str, str],
        Field(
            description="Mapping of custom lookup names to a python importable path "
            "for the class that will be used to resolve the lookups.",
        ),
    ] = {}
    mappings: Annotated[
        dict[str, dict[str, dict[str, Any]]],
        Field(description="Mappings that will be appended to all stack templates."),
    ] = {}
    namespace: Annotated[
        str,
        Field(
            description="The namespace used to prefix stack names to create separation "
            "within an AWS account.",
        ),
    ]
    namespace_delimiter: Annotated[
        str,
        Field(
            description="Character used to separate the namespace and stack name "
            "when the namespace is prepended.",
        ),
    ] = "-"
    package_sources: Annotated[
        CfnginPackageSourcesDefinitionModel,
        Field(
            description="Map of additional package sources to include when "
            "processing this configuration file.",
        ),
    ] = CfnginPackageSourcesDefinitionModel()
    persistent_graph_key: Annotated[
        str | None,
        Field(
            description="Key for an AWS S3 object used to track a graph of stacks "
            "between executions.",
        ),
    ] = None
    post_deploy: Annotated[
        list[CfnginHookDefinitionModel] | dict[str, CfnginHookDefinitionModel],
        Field(title="Post Deploy Hooks"),
    ] = []
    post_destroy: Annotated[
        list[CfnginHookDefinitionModel] | dict[str, CfnginHookDefinitionModel],
        Field(title="Pre Destroy Hooks"),
    ] = []
    pre_deploy: Annotated[
        list[CfnginHookDefinitionModel] | dict[str, CfnginHookDefinitionModel],
        Field(title="Pre Deploy Hooks"),
    ] = []
    pre_destroy: Annotated[
        list[CfnginHookDefinitionModel] | dict[str, CfnginHookDefinitionModel],
        Field(title="Pre Destroy Hooks"),
    ] = []
    service_role: Annotated[
        str | None,
        Field(
            title="Service Role ARN",
            description="Specify an IAM Role for CloudFormation to use.",
        ),
    ] = None
    stacks: Annotated[
        list[CfnginStackDefinitionModel] | dict[str, CfnginStackDefinitionModel],
        Field(
            description="Define CloudFormation stacks using a Blueprint or Template.",
        ),
    ] = []
    sys_path: Annotated[
        Path | None,
        Field(
            title="sys.path",
            description="Path to append to $PATH. This is also the root of relative paths.",
        ),
    ] = None
    tags: Annotated[
        dict[str, str] | None,
        Field(
            description="Tags to try to apply to all resources created from this configuration file.",
        ),
    ] = None  # NOTE (kyle): `None` is significant here
    template_indent: Annotated[
        int,
        Field(
            description="Number of spaces per indentation level to use when "
            "rendering/outputting CloudFormation templates.",
        ),
    ] = 4

    _resolve_path_fields = field_validator("cfngin_cache_dir", "sys_path")(utils.resolve_path_field)

    @field_validator("post_deploy", "post_destroy", "pre_deploy", "pre_destroy", mode="before")
    @classmethod
    def _convert_hook_definitions(
        cls, v: dict[str, Any] | list[dict[str, Any]]
    ) -> list[dict[str, Any]]:
        """Convert hooks defined as a dict to a list."""
        if isinstance(v, list):
            return v
        return list(v.values())

    @field_validator("stacks", mode="before")
    @classmethod
    def _convert_stack_definitions(
        cls, v: dict[str, Any] | list[dict[str, Any]]
    ) -> list[dict[str, Any]]:
        """Convert ``stacks`` defined as a dict to a list."""
        if isinstance(v, list):
            return v
        result: list[dict[str, Any]] = []
        for name, stack in copy.deepcopy(v).items():
            stack["name"] = name
            result.append(stack)
        return result

    @field_validator("stacks")
    @classmethod
    def _validate_unique_stack_names(
        cls, stacks: list[CfnginStackDefinitionModel]
    ) -> list[CfnginStackDefinitionModel]:
        """Validate that each Stack has a unique name."""
        stack_names = [stack.name for stack in stacks]
        if len(set(stack_names)) != len(stack_names):
            for i, name in enumerate(stack_names):
                if stack_names.count(name) != 1:
                    raise ValueError(f"Duplicate stack {name} found at index {i}")
        return stacks

    @classmethod
    def parse_file(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls: type[Self], path: str | Path
    ) -> Self:
        """Parse a file."""
        return cls.model_validate(
            yaml.safe_load(
                Path(path).read_text(encoding=locale.getpreferredencoding(do_setlocale=False))
            )
        )



"""File info builder.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/fileinfobuilder.py

"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .file_info import FileInfo

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    from mypy_boto3_s3.client import S3Client

    from .file_generator import FileStats
    from .parameters import ParametersDataModel


class FileInfoBuilder:
    """Takes a ``FileStats`` object's attributes and generates a ``FileInfo`` object."""

    def __init__(
        self,
        *,
        client: S3Client,
        is_stream: bool = False,
        parameters: ParametersDataModel | None = None,
        source_client: Any | None = None,
    ) -> None:
        """Instantiate class.

        Args:
            client: boto3 S3 client.
            is_stream: If the file is a stream.
            parameters: A dictionary of important values this is assigned in
                the ``BasicTask`` object.
            source_client: Client to handle the source.

        """
        self._client = client
        self._source_client = client
        if source_client is not None:
            self._source_client = source_client
        self._parameters = parameters
        self._is_stream = is_stream

    def call(self, files: Iterable[FileStats]) -> Generator[FileInfo, None, None]:
        """Iterate generator of ``FileStats`` to generate ``FileInfo`` objects."""
        for file_base in files:
            file_info = self._inject_info(file_base)
            yield file_info

    def _inject_info(self, file_base: FileStats) -> FileInfo:
        """Inject info.

        Args:
            file_base: File stats to include in the FileInfo object.

        """
        delete_enabled = self._parameters.delete if self._parameters else False
        return FileInfo(
            **file_base.dict(),  # pyright: ignore[reportArgumentType]
            **(
                {"client": self._source_client, "source_client": self._client}
                if file_base.operation_name == "delete" and delete_enabled
                else {"client": self._client, "source_client": self._source_client}
            ),
        )



"""Lookup to provide a default value."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from ....lookups.handlers.base import LookupHandler

if TYPE_CHECKING:

    from ....context import CfnginContext


class DefaultLookup(LookupHandler["CfnginContext"]):
    """Lookup to provide a default value."""

    TYPE_NAME: ClassVar[str] = "default"
    """Name that the Lookup is registered as."""

    @classmethod
    def handle(cls, value: str, context: CfnginContext | None = None, **_: Any) -> Any:
        """Use a value from the environment or fall back to a default value.

        Allows defaults to be set at the config file level.

        Args:
            value: Parameter(s) given to this lookup. ``<env_var>::<default value>``
            context: Context instance.

        Example:
            ::

                Groups: ${default app_security_groups::sg-12345,sg-67890}

            If ``app_security_groups`` is defined in the environment, its
            defined value will be returned. Otherwise, ``sg-12345,sg-67890``
            will be the returned value.

        """
        try:
            env_var_name, default_val = value.split("::", 1)
        except ValueError:
            raise ValueError(
                f"Invalid value for default: {value}. Must be in "
                "<env_var>::<default value> format."
            ) from None

        if context and env_var_name in context.parameters:
            return context.parameters[env_var_name]
        return default_val



"""Terraform module."""

from __future__ import annotations

import json
import logging
import re
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import hcl
from send2trash import send2trash
from typing_extensions import Literal

from .._logging import PrefixAdaptor
from ..compat import cached_property
from ..config.models.runway.options.terraform import (
    RunwayTerraformBackendConfigDataModel,
    RunwayTerraformModuleOptionsDataModel,
)
from ..env_mgr.tfenv import TFEnvManager
from ..mixins import DelCachedPropMixin
from ..utils import DOC_SITE, Version, which
from .base import ModuleOptions, RunwayModule
from .utils import run_module_command

if TYPE_CHECKING:
    from .._logging import RunwayLogger
    from ..context import RunwayContext
    from ..core.components import DeployEnvironment

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


def gen_workspace_tfvars_files(environment: str, region: str) -> list[str]:
    """Generate possible Terraform workspace tfvars filenames."""
    return [
        # Give preference to explicit environment-region files
        f"{environment}-{region}.tfvars",
        # Fallback to environment name only
        f"{environment}.tfvars",
    ]


def update_env_vars_with_tf_var_values(
    os_env_vars: dict[str, str],
    tf_vars: dict[str, dict[str, Any] | list[Any] | str],
) -> dict[str, str]:
    """Return os_env_vars with TF_VAR values for each tf_var."""
    # https://www.terraform.io/docs/commands/environment-variables.html#tf_var_name
    for key, val in tf_vars.items():
        if isinstance(val, dict):
            value = ", ".join(
                nestedkey + ' = "' + str(nestedval) + '"' for (nestedkey, nestedval) in val.items()
            )
            os_env_vars[f"TF_VAR_{key}"] = f"{{ {value} }}"
        elif isinstance(val, list):
            os_env_vars[f"TF_VAR_{key}"] = str(json.dumps(val))
        else:
            os_env_vars[f"TF_VAR_{key}"] = str(val)
    return os_env_vars


class TerraformBackendConfig(ModuleOptions):
    """Terraform backend configuration module options."""

    def __init__(
        self,
        data: RunwayTerraformBackendConfigDataModel,
        deploy_environment: DeployEnvironment,
        path: Path,
    ) -> None:
        """Instantiate class.

        Args:
            data: Options parsed into a data model.
            deploy_environment: Current deploy environment.
            path: Module path.

        """
        self.bucket = data.bucket
        self.data = data
        self.dynamodb_table = data.dynamodb_table
        self.env = deploy_environment
        self.path = path
        if data and not data.region:
            data.region = deploy_environment.aws_region  # default to region from env
        self.region = data.region

    @cached_property
    def config_file(self) -> Path | None:
        """Backend configuration file."""
        return self.get_backend_file(self.path, self.env.name, self.env.aws_region)

    @cached_property
    def init_args(self) -> list[str]:
        """Return command line arguments for init."""
        result: list[str] = []
        for k, v in self.data.model_dump(exclude_none=True).items():
            result.extend(["-backend-config", f"{k}={v}"])
        if not result:
            if self.config_file:
                LOGGER.verbose("using backend config file: %s", self.config_file.name)
                return [f"-backend-config={self.config_file.name}"]
            LOGGER.info(
                "backend file not found -- looking for one of: %s",
                ", ".join(self.gen_backend_filenames(self.env.name, self.env.aws_region)),
            )
            return []
        LOGGER.info("using backend values from runway.yml")
        LOGGER.debug("provided backend values: %s", json.dumps(result))
        return result

    def get_full_configuration(self) -> dict[str, str]:
        """Get full backend configuration."""
        if not self.config_file:
            return self.data.model_dump(exclude_none=True)
        result = cast(dict[str, str], hcl.loads(self.config_file.read_text()))
        result.update(self.data.model_dump(exclude_none=True))
        return result

    @classmethod
    def get_backend_file(cls, path: Path, environment: str, region: str) -> Path | None:
        """Determine Terraform backend file.

        Args:
            path: Path to the module.
            environment: Current deploy environment.
            region: Current AWS region.

        """
        backend_filenames = cls.gen_backend_filenames(environment, region)
        for name in backend_filenames:
            test_path = path / name
            if test_path.is_file():
                return test_path
        return None

    @staticmethod
    def gen_backend_filenames(environment: str, region: str) -> list[str]:
        """Generate possible Terraform backend filenames.

        Args:
            environment: Current deploy environment.
            region : Current AWS region.

        """
        formats = [
            "backend-{environment}-{region}.{extension}",
            "backend-{environment}.{extension}",
            "backend-{region}.{extension}",
            "backend.{extension}",
        ]
        result: list[str] = []
        for fmt in formats:
            for ext in ["hcl", "tfvars"]:
                result.append(  # noqa: PERF401
                    fmt.format(environment=environment, extension=ext, region=region)
                )
        return result

    @classmethod
    def parse_obj(
        cls,
        deploy_environment: DeployEnvironment,
        obj: object,
        path: Path | None = None,
    ) -> TerraformBackendConfig:
        """Parse options definition and return an options object.

        Args:
            deploy_environment: Current deploy environment.
            obj: Object to parse.
            path: Module path.

        """
        return cls(
            data=RunwayTerraformBackendConfigDataModel.model_validate(obj),
            deploy_environment=deploy_environment,
            path=path or Path.cwd(),
        )


class TerraformOptions(ModuleOptions):
    """Module options for Terraform.

    Attributes:
        args: CLI arguments/options to pass to Terraform.
        data: Options parsed into a data model.
        env: Current deploy environment.
        path: Module path.
        version: String containing a Terraform version.
        write_auto_tfvars: Optionally write parameters to a tfvars file instead
            of updating variables.

    """

    def __init__(
        self,
        data: RunwayTerraformModuleOptionsDataModel,
        deploy_environment: DeployEnvironment,
        path: Path | None = None,
    ) -> None:
        """Instantiate class.

        Args:
            deploy_environment: Current deploy environment.
            data: Options parsed into a data model.
            path: Module path.

        """
        self.args = data.args
        self.data = data
        self.env = deploy_environment
        self.path = path or Path.cwd()
        self.version = data.version
        self.workspace = data.workspace or deploy_environment.name
        self.write_auto_tfvars = data.write_auto_tfvars

    @cached_property
    def backend_config(self) -> TerraformBackendConfig:
        """Backend configuration options."""
        return TerraformBackendConfig.parse_obj(
            deploy_environment=self.env,
            obj=self.data.backend_config or {},
            path=self.path,
        )

    @classmethod
    def parse_obj(
        cls,
        deploy_environment: DeployEnvironment,
        obj: object,
        path: Path | None = None,
    ) -> TerraformOptions:
        """Parse options definition and return an options object.

        Args:
            deploy_environment: Current deploy environment.
            obj: Object to parse.
            path: Module path.

        """
        return cls(
            data=RunwayTerraformModuleOptionsDataModel.model_validate(obj),
            deploy_environment=deploy_environment,
            path=path or Path.cwd(),
        )


TerraformActionTypeDef = Literal[
    "apply",
    "destroy",
    "get",
    "init",
    "plan",
    "workspace_list",
    "workspace_new",
    "workspace_select",
    "workspace_show",
]


class Terraform(RunwayModule[TerraformOptions], DelCachedPropMixin):
    """Terraform Runway Module."""

    def __init__(
        self,
        context: RunwayContext,
        *,
        explicitly_enabled: bool | None = False,
        logger: RunwayLogger = LOGGER,
        module_root: Path,
        name: str | None = None,
        options: dict[str, Any] | ModuleOptions | None = None,
        parameters: dict[str, Any] | None = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object for the current session.
            explicitly_enabled: Whether or not the module is explicitly enabled.
                This is can be set in the event that the current environment being
                deployed to matches the defined environments of the module/deployment.
            logger: Used to write logs.
            module_root: Root path of the module.
            name: Name of the module.
            options: Options passed to the module class from the config as ``options``
                or ``module_options`` if coming from the deployment level.
            parameters: Values to pass to the underlying infrastructure as code
                tool that will alter the resulting infrastructure being deployed.
                Used to templatize IaC.

        """
        super().__init__(
            context,
            explicitly_enabled=explicitly_enabled,
            logger=logger,
            module_root=module_root,
            name=name,
            options=TerraformOptions.parse_obj(
                deploy_environment=context.env, obj=options or {}, path=module_root
            ),
            parameters=parameters,
        )
        # logger needs to be created here to use the correct logger
        self.logger = PrefixAdaptor(self.name, logger)
        self.required_workspace = self.options.workspace

    @cached_property
    def auto_tfvars(self) -> Path:
        """Return auto.tfvars file if one is being used."""
        file_path = self.path / "runway-parameters.auto.tfvars.json"
        if self.parameters and self.options.write_auto_tfvars:
            if self.version < Version("0.10"):
                self.logger.warning(
                    "Terraform version does not support the use of "
                    "*.auto.tfvars; some variables may be missing"
                )
            file_path.write_text(json.dumps(self.parameters, indent=4))
        return file_path

    @cached_property
    def current_workspace(self) -> str:
        """Wrap "terraform_workspace_show" to cache the value.

        Returns:
            The currently active Terraform workspace.

        """
        return self.terraform_workspace_show()

    @cached_property
    def env_file(self) -> list[str]:
        """Find the environment file for the module."""
        result: list[str] = []
        for name in gen_workspace_tfvars_files(self.ctx.env.name, self.ctx.env.aws_region):
            test_path = self.path / name
            if test_path.is_file():
                result.append("-var-file=" + test_path.name)
                break  # stop looking if one is found
        return result

    @property
    def skip(self) -> bool:
        """Determine if the module should be skipped."""
        if self.parameters or self.env_file:
            return False
        self.logger.info(
            "skipped; tfvars file for this environment/region not found "
            "and no parameters provided -- looking for one of: %s",
            ", ".join(gen_workspace_tfvars_files(self.ctx.env.name, self.ctx.env.aws_region)),
        )
        return True

    @cached_property
    def tfenv(self) -> TFEnvManager:
        """Terraform environment manager."""
        return TFEnvManager(self.path)

    @cached_property
    def tf_bin(self) -> str:
        """Path to Terraform binary."""
        try:
            return self.tfenv.install(self.options.version)
        except ValueError:
            self.logger.debug("terraform install failed", exc_info=True)
            self.logger.verbose("terraform version not specified; resorting to global install")
            if which("terraform"):
                return "terraform"
        self.logger.error("terraform not available and a version to install not specified")
        self.logger.error(
            "learn how to use Runway to manage Terraform versions at "
            "%s/page/terraform/advanced_features.html#version-management",
            DOC_SITE,
        )
        sys.exit(1)

    @cached_property
    def version(self) -> Version:
        """Version of Terraform being used."""
        if not self.tfenv.current_version and self.options.version:
            self.tfenv.set_version(self.options.version)
        if self.tfenv.version:
            return self.tfenv.version
        version = self.tfenv.get_version_from_executable(self.tf_bin)
        if version:
            return version
        raise ValueError(f"unable to retrieve version from {self.tf_bin}")

    def cleanup_dot_terraform(self) -> None:
        """Remove .terraform excluding the plugins directly.

        This step is crucial for allowing Runway to deploy to multiple regions
        or deploy environments without prompting the user for input.

        The plugins directory is retained to improve performance when they
        are used by subsequent runs.

        """
        dot_terraform = self.path / ".terraform"
        if not dot_terraform.is_dir():
            self.logger.debug(".terraform directory does not exist; skipped cleanup")
            return

        self.logger.verbose(
            ".terraform directory exists from a previous run; removing some of its contents"
        )
        for child in dot_terraform.iterdir():
            if child.name == "plugins" and child.is_dir():
                self.logger.debug("directory retained: %s", child)
                continue
            self.logger.debug("removing: %s", child)
            send2trash(str(child))  # does not support Path objects

    def deploy(self) -> None:
        """Run Terraform apply."""
        self.run("apply")

    def destroy(self) -> None:
        """Run Terraform destroy."""
        self.run("destroy")

    def gen_command(
        self, command: list[str] | str | tuple[str, ...], args_list: list[str] | None = None
    ) -> list[str]:
        """Generate Terraform command."""
        cmd = (
            [self.tf_bin, *command]
            if isinstance(command, (list, tuple))
            else [self.tf_bin, command]
        )
        cmd.extend(args_list or [])
        if self.ctx.no_color:
            cmd.append("-no-color")
        return cmd

    def handle_backend(self) -> None:
        """Handle backend configuration.

        This needs to be run before "skip" is assessed or env_file/auto_tfvars
        is used in case their behavior needs to be altered.

        """
        if not self.tfenv.backend["type"]:
            self.logger.info(
                "unable to determine backend for module; no special handling will be applied"
            )
            return
        handler = f"_{self.tfenv.backend['type']}_backend_handler"
        if hasattr(self, handler):
            self.tfenv.backend["config"].update(
                self.options.backend_config.get_full_configuration()
            )
            self.logger.debug("full backend config: %s", json.dumps(self.tfenv.backend["config"]))
            self.logger.verbose("handling use of backend config: %s", self.tfenv.backend["type"])
            self[f"_{self.tfenv.backend['type']}_backend_handler"]()
        else:
            self.logger.verbose(
                'backed "%s" does not require special handling',
                self.tfenv.backend["type"],
                exc_info=True,
            )

    def _remote_backend_handler(self) -> None:
        """Handle special setting required for using a remote backend."""
        if not self.tfenv.backend["config"].get("workspaces"):
            self.logger.warning(
                '"workspaces" not defined in backend config; unable to '
                "apply appropriate handling -- processing may fail"
            )
            return

        self.logger.verbose(
            "forcing parameters to be written to runway-parameters.auto.tfvars.json"
        )
        # this is because variables cannot be added inline or via environment
        # variables when using a remote backend
        self.options.write_auto_tfvars = True

        if self.tfenv.backend["config"]["workspaces"].get("prefix"):
            self.logger.verbose("handling use of backend config: remote.workspaces.prefix")
            self.ctx.env.vars.update({"TF_WORKSPACE": self.ctx.env.name})
            self.logger.verbose(
                'set environment variable "TF_WORKSPACE" to avoid prompt '
                "during init by pre-selecting an appropriate workspace"
            )

        if self.tfenv.backend["config"]["workspaces"].get("name"):
            self.logger.verbose("handling use of backend config: remote.workspaces.name")
            # this can't be set or it will cause errors
            self.ctx.env.vars.pop("TF_WORKSPACE", None)
            self.required_workspace = "default"
            self.logger.info(
                'forcing use of static workspace "default"; '
                'required for use of "backend.remote.workspaces.name"'
            )

    def handle_parameters(self) -> None:
        """Handle parameters.

        Either updating environment variables or writing to a file.

        """
        if self.auto_tfvars.exists():
            return

        self.ctx.env.vars = update_env_vars_with_tf_var_values(self.ctx.env.vars, self.parameters)

    def init(self) -> None:
        """Run init."""
        self.run("init")

    def plan(self) -> None:
        """Run Terraform plan."""
        self.run("plan")

    def terraform_apply(self) -> None:
        """Execute ``terraform apply`` command.

        https://www.terraform.io/docs/cli/commands/apply.html

        """
        args_list = self.env_file + self.options.args.apply
        if self.ctx.env.ci:
            args_list.append("-auto-approve=true")
        else:
            args_list.append("-auto-approve=false")
        run_module_command(
            self.gen_command("apply", args_list),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )

    def terraform_destroy(self) -> None:
        """Execute ``terraform destroy`` command.

        https://www.terraform.io/docs/cli/commands/destroy.html

        """
        if self.version >= Version("0.15.2"):
            return self._terraform_destroy_15_2()
        if self.version >= Version("0.12"):
            return self._terraform_destroy_12()
        return self._terraform_destroy_legacy()

    def _terraform_destroy_12(self) -> None:
        """Execute ``terraform destroy -auto-approve`` command.

        Compatible with Terraform >=0.12.0, <0.15.2.

        """
        return run_module_command(
            self.gen_command("destroy", ["-auto-approve", *self.env_file]),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )

    def _terraform_destroy_15_2(self) -> None:
        """Execute ``terraform apply -destroy -auto-approve`` command.

        Compatible with Terraform >=0.15.2.

        """
        return run_module_command(
            self.gen_command("apply", ["-destroy", "-auto-approve", *self.env_file]),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )

    def _terraform_destroy_legacy(self) -> None:
        """Execute ``terraform destroy -force`` command.

        Compatible with Terraform <0.12.0.

        """
        return run_module_command(
            self.gen_command("destroy", ["-force", *self.env_file]),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )

    def terraform_get(self) -> None:
        """Execute ``terraform get`` command.

        https://www.terraform.io/docs/cli/commands/get.html

        """
        self.logger.info("downloading and updating Terraform modules")
        run_module_command(
            self.gen_command("get", ["-update=true"]),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )

    def terraform_init(self) -> None:
        """Execute ``terraform init`` command.

        https://www.terraform.io/docs/cli/commands/init.html

        """
        cmd = self.gen_command(
            "init",
            ["-reconfigure", *self.options.backend_config.init_args, *self.options.args.init],
        )
        try:
            run_module_command(
                cmd,
                env_vars=self.ctx.env.vars,
                exit_on_error=False,
                logger=self.logger,
            )
        except subprocess.CalledProcessError as shelloutexc:
            # cleaner output by not letting the exception raise
            sys.exit(shelloutexc.returncode)

    def terraform_plan(self) -> None:
        """Execute ``terraform plan`` command.

        https://www.terraform.io/docs/cli/commands/plan.html

        """
        run_module_command(
            self.gen_command("plan", self.env_file + self.options.args.plan),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )

    def terraform_workspace_list(self) -> str:
        """Execute ``terraform workspace list`` command.

        https://www.terraform.io/docs/cli/commands/workspace/list.html

        Returns:
            str: The available Terraform workspaces.

        """
        self.logger.debug("listing available Terraform workspaces")
        workspaces = subprocess.check_output(
            self.gen_command(["workspace", "list"]), env=self.ctx.env.vars
        ).decode()
        self.logger.debug("available Terraform workspaces:\n%s", workspaces)
        return workspaces

    def terraform_workspace_new(self, workspace: str) -> None:
        """Execute ``terraform workspace new`` command.

        permanently to https://www.terraform.io/docs/cli/commands/workspace/new.html

        Args:
            workspace: Terraform workspace to create.

        """
        self.logger.debug("creating workspace: %s", workspace)
        run_module_command(
            self.gen_command(["workspace", "new"], [workspace]),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )
        self.logger.debug("workspace created")

    def terraform_workspace_select(self, workspace: str) -> None:
        """Execute ``terraform workspace select`` command.

        https://www.terraform.io/docs/cli/commands/workspace/select.html

        Args:
            workspace: Terraform workspace to select.

        """
        self.logger.debug(
            'switching Terraform workspace from "%s" to "%s"',
            self.current_workspace,
            workspace,
        )
        run_module_command(
            self.gen_command(["workspace", "select"], [workspace]),
            env_vars=self.ctx.env.vars,
            logger=self.logger,
        )
        self._del_cached_property("current_workspace")

    def terraform_workspace_show(self) -> str:
        """Execute ``terraform workspace show`` command.

        https://www.terraform.io/docs/cli/commands/workspace/show.html

        Returns:
            The current Terraform workspace.

        """
        self.logger.debug("using Terraform to get the current workspace")
        workspace = (
            subprocess.check_output(self.gen_command(["workspace", "show"]), env=self.ctx.env.vars)
            .strip()
            .decode()
        )
        self.logger.debug("current Terraform workspace: %s", workspace)
        return workspace

    def run(self, action: TerraformActionTypeDef) -> None:
        """Run module."""
        try:
            self.handle_backend()
            if self.skip:
                return
            self.cleanup_dot_terraform()
            self.handle_parameters()
            self.logger.info("init (in progress)")
            self.terraform_init()
            if self.current_workspace != self.required_workspace:
                if re.compile(f"^[*\\s]\\s{self.required_workspace}$", re.MULTILINE).search(
                    self.terraform_workspace_list()
                ):
                    self.terraform_workspace_select(self.required_workspace)
                else:
                    self.terraform_workspace_new(self.required_workspace)
                self.logger.verbose("re-running init after workspace change...")
                self.terraform_init()
            self.terraform_get()
            self.logger.info("init (complete)")
            if action != "init":
                self.logger.info("%s (in progress)", action)
                self["terraform_" + action]()
                self.logger.info("%s (complete)", action)
        finally:
            if self.auto_tfvars.exists():
                self.auto_tfvars.unlink()



"""Uninstall Terraform version(s) that were installed by Runway and/or tfenv."""

# docs: file://./../../../../docs/source/commands.rst
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, cast

import click

from ....env_mgr.tfenv import TFEnvManager
from ... import options

if TYPE_CHECKING:
    from runway._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__.replace("._", ".")))


@click.command("uninstall", short_help="uninstall terraform")
@click.argument("version", metavar="[<version>]", default=None, required=False)
@click.option(
    "--all",
    "all_versions",
    default=False,
    help="Uninstall all versions of Terraform.",
    is_flag=True,
    required=False,
    show_default=True,
)
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def uninstall(
    ctx: click.Context,
    *,
    version: str | None = None,
    all_versions: bool = False,
    **_: Any,
) -> None:
    """Uninstall the specified <version> of Terraform (e.g. 0.12.0) or all installed versions.

    If no version is specified, Runway will attempt to find and read a
    ".terraform-version" file in the current directory.

    """
    tfenv = TFEnvManager()
    version = version or (str(tfenv.version) if tfenv.version else None)
    if version and not all_versions:
        if not tfenv.uninstall(version):
            ctx.exit(1)
        return
    if all_versions:
        LOGGER.notice("uninstalling all versions of Terraform...")
        for v in tfenv.list_installed():
            tfenv.uninstall(v.name)
        LOGGER.success("all versions of Terraform have been uninstalled")
        return
    LOGGER.error("version not specified")
    ctx.exit(1)
